#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Volume V3 Volume action implementations"""

import argparse
from collections.abc import Iterable, Sequence
import functools
import logging
from typing import Any

from cliff import columns as cliff_columns
from openstack.block_storage.v3 import volume as _volume
from openstack import exceptions as sdk_exceptions
from openstack import utils as sdk_utils
from osc_lib.cli import format_columns
from osc_lib.cli import parseractions
from osc_lib import exceptions
from osc_lib import utils

from openstackclient.api import volume_v3
from openstackclient import command
from openstackclient.common import pagination
from openstackclient.i18n import _
from openstackclient.identity import common as identity_common


LOG = logging.getLogger(__name__)


class KeyValueHintAction(argparse.Action):
    """Uses KeyValueAction or KeyValueAppendAction based on the given key"""

    APPEND_KEYS = ('same_host', 'different_host')

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self._key_value_action = parseractions.KeyValueAction(*args, **kwargs)
        self._key_value_append_action = parseractions.KeyValueAppendAction(
            *args, **kwargs
        )
        super().__init__(*args, **kwargs)

    def __call__(
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: Any,
        option_string: str | None = None,
    ) -> None:
        if values.startswith(self.APPEND_KEYS):
            self._key_value_append_action(
                parser, namespace, values, option_string=option_string
            )
        else:
            self._key_value_action(
                parser, namespace, values, option_string=option_string
            )


class AttachmentsColumn(cliff_columns.FormattableColumn[list[Any]]):
    """Formattable column for attachments column.

    Unlike the parent FormattableColumn class, the initializer of the
    class takes server_cache as the second argument.
    osc_lib.utils.get_item_properties instantiate cliff FormattableColumn
    object with a single parameter "column value", so you need to pass
    a partially initialized class like
    ``functools.partial(AttachmentsColumn, server_cache)``.
    """

    def __init__(
        self, value: list[Any], server_cache: dict[str, Any] | None = None
    ) -> None:
        super().__init__(value)
        self._server_cache = server_cache or {}

    def human_readable(self) -> str:
        """Return a formatted string of a volume's attached instances

        :rtype: a string of formatted instances
        """

        msg = ''
        for attachment in self._value:
            server = attachment['server_id']
            if server in self._server_cache.keys():
                server = self._server_cache[server].name
            device = attachment['device']
            msg += f'Attached to {server} on {device} '
        return msg


def _format_volume(volume: _volume.Volume) -> dict[str, object]:
    # Some columns returned by openstacksdk should not be shown because they're
    # either irrelevant or duplicates
    ignored_columns = {
        # computed columns
        'location',
        # create-only columns
        'OS-SCH-HNT:scheduler_hints',
        'imageRef',
        # removed columns
        'os-volume-replication:driver_data',
        'os-volume-replication:extended_status',
        # unnecessary columns
        'links',
    }
    optional_columns = {
        # only present if part of a consistency group
        'consistencygroup_id',
        # only present if the volume is encrypted
        'encryption_key_id',
        # only present if there are image properties associated
        'volume_image_metadata',
    }

    info = volume.to_dict(original_names=True)
    data = {}
    for key, value in info.items():
        if key in ignored_columns:
            continue

        if key in optional_columns:
            if info[key] is None:
                continue

        data[key] = value

    data.update(
        {
            'properties': format_columns.DictColumn(data.pop('metadata')),
            'type': data.pop('volume_type'),
        }
    )

    return data


class CreateVolume(command.ShowOne):
    _description = _("Create new volume")

    @staticmethod
    def _check_size_arg(args: argparse.Namespace) -> None:
        """Check whether --size option is required or not.

        Require size parameter in case if any of the following is not
        specified:

        * snapshot
        * source volume
        * backup
        * remote source (volume to be managed)
        """

        if (
            args.snapshot or args.source or args.backup or args.remote_source
        ) is None and args.size is None:
            msg = _(
                "--size is a required option if none of --snapshot, "
                "--backup, --source, or --remote-source are provided."
            )
            raise exceptions.CommandError(msg)

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "name",
            metavar="<name>",
            nargs="?",
            help=_("Volume name"),
        )
        parser.add_argument(
            "--size",
            metavar="<size>",
            type=int,
            help=_(
                "Volume size in GB (required unless --snapshot or "
                "--source specified)"
            ),
        )
        parser.add_argument(
            "--type",
            metavar="<volume-type>",
            help=_("Set the type of volume"),
        )
        source_group = parser.add_mutually_exclusive_group()
        source_group.add_argument(
            "--image",
            metavar="<image>",
            help=_("Use <image> as source of volume (name or ID)"),
        )
        source_group.add_argument(
            "--snapshot",
            metavar="<snapshot>",
            help=_("Use <snapshot> as source of volume (name or ID)"),
        )
        source_group.add_argument(
            "--source",
            metavar="<volume>",
            help=_("Volume to clone (name or ID)"),
        )
        source_group.add_argument(
            "--source-replicated",
            metavar="<replicated-volume>",
            help=argparse.SUPPRESS,
        )
        source_group.add_argument(
            "--backup",
            metavar="<backup>",
            help=_(
                "Restore backup to a volume (name or ID) "
                "(supported by --os-volume-api-version 3.47 or later)"
            ),
        )
        source_group.add_argument(
            "--remote-source",
            metavar="<key=value>",
            action=parseractions.KeyValueAction,
            help=_(
                "The attribute(s) of the existing remote volume "
                "(admin required) (repeat option to specify multiple "
                "attributes, e.g.: '--remote-source source-name=test_name "
                "--remote-source source-id=test_id')"
            ),
        )
        parser.add_argument(
            "--description",
            metavar="<description>",
            help=_("Volume description"),
        )
        parser.add_argument(
            "--availability-zone",
            metavar="<availability-zone>",
            help=_("Create volume in <availability-zone>"),
        )
        parser.add_argument(
            "--consistency-group",
            metavar="consistency-group>",
            help=_("Consistency group where the new volume belongs to"),
        )
        parser.add_argument(
            "--property",
            metavar="<key=value>",
            action=parseractions.KeyValueAction,
            dest="properties",
            help=_(
                "Set a property to this volume "
                "(repeat option to set multiple properties)"
            ),
        )
        parser.add_argument(
            "--hint",
            metavar="<key=value>",
            action=KeyValueHintAction,
            help=_(
                "Arbitrary scheduler hint key-value pairs to help creating "
                "a volume. Repeat the option to set multiple hints. "
                "'same_host' and 'different_host' get values appended when "
                "repeated, all other keys take the last given value"
            ),
        )
        bootable_group = parser.add_mutually_exclusive_group()
        bootable_group.add_argument(
            "--bootable",
            action="store_true",
            dest="bootable",
            default=None,
            help=_("Mark volume as bootable"),
        )
        bootable_group.add_argument(
            "--non-bootable",
            action="store_false",
            dest="bootable",
            default=None,
            help=_("Mark volume as non-bootable (default)"),
        )
        readonly_group = parser.add_mutually_exclusive_group()
        readonly_group.add_argument(
            "--read-only",
            action="store_true",
            dest="read_only",
            default=None,
            help=_("Set volume to read-only access mode"),
        )
        readonly_group.add_argument(
            "--read-write",
            action="store_false",
            dest="read_only",
            default=None,
            help=_("Set volume to read-write access mode (default)"),
        )
        parser.add_argument(
            "--host",
            metavar="<host>",
            help=_(
                "Cinder host on which the existing volume resides; "
                "takes the form: host@backend-name#pool. This is only "
                "used along with the --remote-source option."
            ),
        )
        parser.add_argument(
            "--cluster",
            metavar="<cluster>",
            help=_(
                "Cinder cluster on which the existing volume resides; "
                "takes the form: cluster@backend-name#pool. This is only "
                "used along with the --remote-source option. "
                "(supported by --os-volume-api-version 3.16 or above)"
            ),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        self._check_size_arg(parsed_args)
        # size is validated in the above call to
        # _check_size_arg where we check that size
        # should be passed if we are not creating a
        # volume from snapshot, backup or source volume
        size = parsed_args.size

        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )
        image_client = self.app.client_manager.image

        if (
            parsed_args.host or parsed_args.cluster
        ) and not parsed_args.remote_source:
            msg = _(
                "The --host and --cluster options are only supported "
                "with --remote-source parameter."
            )
            raise exceptions.CommandError(msg)

        if parsed_args.backup and not sdk_utils.supports_microversion(
            volume_client, '3.47'
        ):
            msg = _(
                "--os-volume-api-version 3.47 or greater is required "
                "to create a volume from backup."
            )
            raise exceptions.CommandError(msg)

        if parsed_args.remote_source:
            if (
                parsed_args.size
                or parsed_args.consistency_group
                or parsed_args.hint
                or parsed_args.read_only is not None
            ):
                msg = _(
                    "The --size, --consistency-group, --hint, --read-only "
                    "and --read-write options are not supported with the "
                    "--remote-source parameter."
                )
                raise exceptions.CommandError(msg)
            if parsed_args.cluster:
                if not sdk_utils.supports_microversion(volume_client, '3.16'):
                    msg = _(
                        "--os-volume-api-version 3.16 or greater is required "
                        "to support the cluster parameter."
                    )
                    raise exceptions.CommandError(msg)
            if parsed_args.cluster and parsed_args.host:
                msg = _(
                    "Only one of --host or --cluster needs to be specified "
                    "to manage a volume."
                )
                raise exceptions.CommandError(msg)
            if not parsed_args.cluster and not parsed_args.host:
                msg = _(
                    "One of --host or --cluster needs to be specified to "
                    "manage a volume."
                )
                raise exceptions.CommandError(msg)
            volume = volume_client.manage_volume(
                host=parsed_args.host,
                cluster=parsed_args.cluster,
                ref=parsed_args.remote_source,
                name=parsed_args.name,
                description=parsed_args.description,
                volume_type=parsed_args.type,
                availability_zone=parsed_args.availability_zone,
                metadata=parsed_args.properties,
                bootable=parsed_args.bootable,
            )
            data = _format_volume(volume)
            col_headers, col_data = zip(*sorted(data.items()))
            return col_headers, col_data

        source_volume = None
        if parsed_args.source:
            source_volume_obj = volume_client.find_volume(
                parsed_args.source, ignore_missing=False
            )
            source_volume = source_volume_obj.id
            size = max(size or 0, source_volume_obj.size)

        consistency_group = None
        if parsed_args.consistency_group:
            consistency_group = volume_v3.find_consistency_group(
                volume_client, parsed_args.consistency_group
            )['id']

        image = None
        if parsed_args.image:
            image = image_client.find_image(
                parsed_args.image, ignore_missing=False
            ).id

        snapshot = None
        if parsed_args.snapshot:
            snapshot_obj = volume_client.find_snapshot(
                parsed_args.snapshot, ignore_missing=False
            )
            snapshot = snapshot_obj.id
            # Cinder requires a value for size when creating a volume
            # even if creating from a snapshot. Cinder will create the
            # volume with at least the same size as the snapshot anyway,
            # so since we have the object here, just override the size
            # value if it's either not given or is smaller than the
            # snapshot size.
            size = max(size or 0, snapshot_obj.size)

        backup = None
        if parsed_args.backup:
            backup_obj = volume_client.find_backup(
                parsed_args.backup, ignore_missing=False
            )
            backup = backup_obj.id
            # As above
            size = max(size or 0, backup_obj.size)

        volume = volume_client.create_volume(
            size=size,
            snapshot_id=snapshot,
            name=parsed_args.name,
            description=parsed_args.description,
            volume_type=parsed_args.type,
            availability_zone=parsed_args.availability_zone,
            metadata=parsed_args.properties,
            image_id=image,
            source_volume_id=source_volume,
            consistency_group_id=consistency_group,
            scheduler_hints=parsed_args.hint,
            backup_id=backup,
        )

        if parsed_args.bootable is not None:
            try:
                if utils.wait_for_status(
                    volume_client.get_volume,
                    volume.id,
                    success_status=['available'],
                    error_status=['error'],
                    sleep_time=1,
                ):
                    volume_client.set_volume_bootable_status(
                        volume, parsed_args.bootable
                    )
                else:
                    msg = _(
                        "Volume status is not available for setting boot state"
                    )
                    raise exceptions.CommandError(msg)
            except Exception as e:
                LOG.error(_("Failed to set volume bootable property: %s"), e)

        if parsed_args.read_only is not None:
            try:
                if utils.wait_for_status(
                    volume_client.get_volume,
                    volume.id,
                    success_status=['available'],
                    error_status=['error'],
                    sleep_time=1,
                ):
                    volume_client.set_volume_readonly(
                        volume, parsed_args.read_only
                    )
                else:
                    msg = _(
                        "Volume status is not available for setting it"
                        "read only."
                    )
                    raise exceptions.CommandError(msg)
            except Exception as e:
                LOG.error(
                    _("Failed to set volume read-only access mode flag: %s"),
                    e,
                )

        data = _format_volume(volume)
        col_headers, col_data = zip(*sorted(data.items()))
        return col_headers, col_data


class DeleteVolume(command.Command):
    _description = _("Delete volume(s)")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "volumes",
            metavar="<volume>",
            nargs="+",
            help=_("Volume(s) to delete (name or ID)"),
        )
        group = parser.add_mutually_exclusive_group()
        group.add_argument(
            "--force",
            action="store_true",
            help=_(
                "Attempt forced removal of volume(s), regardless of state "
                "(defaults to False)"
            ),
        )
        group.add_argument(
            "--cascade",
            action="store_true",
            help=_(
                "Remove any snapshots along with volume(s) (defaults to False)"
            ),
        )
        group.add_argument(
            # now called "cascade", accept old arg for compatibility
            "--purge",
            action="store_true",
            help=argparse.SUPPRESS,
            dest='cascade',
        )
        parser.add_argument(
            '--remote',
            action='store_true',
            help=_("Specify this parameter to unmanage a volume."),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )
        result = 0

        if parsed_args.remote and (parsed_args.force or parsed_args.cascade):
            msg = _(
                "The --force and --cascade options are not "
                "supported with the --remote parameter."
            )
            raise exceptions.CommandError(msg)

        for volume in parsed_args.volumes:
            try:
                volume_obj = volume_client.find_volume(
                    volume, ignore_missing=False
                )
                if parsed_args.remote:
                    volume_client.unmanage_volume(volume_obj.id)
                else:
                    volume_client.delete_volume(
                        volume_obj.id,
                        force=parsed_args.force,
                        cascade=parsed_args.cascade,
                    )
            except Exception as e:
                result += 1
                LOG.error(
                    _(
                        "Failed to delete volume with "
                        "name or ID '%(volume)s': %(e)s"
                    ),
                    {'volume': volume, 'e': e},
                )

        if result > 0:
            total = len(parsed_args.volumes)
            msg = _("%(result)s of %(total)s volumes failed to delete.") % {
                'result': result,
                'total': total,
            }
            raise exceptions.CommandError(msg)


class ListVolume(command.Lister):
    _description = _("List volumes")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--project',
            metavar='<project>',
            help=_('Filter results by project (name or ID) (admin only)'),
        )
        identity_common.add_project_domain_option_to_parser(parser)
        parser.add_argument(
            '--user',
            metavar='<user>',
            help=_('Filter results by user (name or ID) (admin only)'),
        )
        identity_common.add_user_domain_option_to_parser(parser)
        parser.add_argument(
            '--name',
            metavar='<name>',
            help=_('Filter results by volume name'),
        )
        parser.add_argument(
            '--status',
            metavar='<status>',
            help=_('Filter results by status'),
        )
        parser.add_argument(
            '--property',
            metavar='<key=value>',
            action=parseractions.KeyValueAction,
            dest='properties',
            help=_(
                'Filter by a property on the volume list '
                '(repeat option to filter by multiple properties) '
            ),
        )
        parser.add_argument(
            '--all-projects',
            action='store_true',
            default=False,
            help=_('Include all projects (admin only)'),
        )
        parser.add_argument(
            '--long',
            action='store_true',
            default=False,
            help=_('List additional fields in output'),
        )
        pagination.add_marker_pagination_option_to_parser(
            parser, include_max_items=False
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[tuple[Any, ...]]]:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )
        identity_client = sdk_utils.ensure_service_version(
            self.app.client_manager.sdk_connection.identity, '3'
        )

        columns: tuple[str, ...] = ('id', 'name', 'status', 'size')
        column_headers: tuple[str, ...] = ('ID', 'Name', 'Status', 'Size')
        if parsed_args.long:
            columns += ('volume_type', 'is_bootable')
            column_headers += ('Type', 'Bootable')
        columns += ('attachments',)
        column_headers += ('Attached to',)
        if parsed_args.long:
            columns += ('metadata',)
            column_headers += ('Properties',)

        kwargs = {}

        if parsed_args.name:
            kwargs['name'] = parsed_args.name

        if parsed_args.status:
            kwargs['status'] = parsed_args.status

        if parsed_args.limit:
            kwargs['limit'] = parsed_args.limit

        if parsed_args.marker:
            kwargs['marker'] = parsed_args.marker

        if parsed_args.project:
            project_id = identity_common.find_project_id_sdk(
                identity_client,
                parsed_args.project,
                parsed_args.project_domain,
            )
            kwargs['project_id'] = project_id

        user_id = None
        if parsed_args.user:
            user_id = identity_common.find_user_id_sdk(
                identity_client, parsed_args.user, parsed_args.user_domain
            )
            kwargs['user_id'] = user_id

        if parsed_args.properties:
            # yes, cinder really does expect a stringified dict...
            kwargs['properties'] = str(parsed_args.properties)

        # set value of 'all_tenants' when using project option
        all_projects = bool(parsed_args.project) or parsed_args.all_projects
        kwargs['all_projects'] = all_projects

        data = list(volume_client.volumes(**kwargs))

        do_server_list = False

        for vol in data:
            if vol.status == 'in-use':
                do_server_list = True
                break

        # Cache the server list
        server_cache = {}
        if do_server_list:
            try:
                compute_client = self.app.client_manager.compute
                for s in compute_client.servers():
                    server_cache[s.id] = s
            except sdk_exceptions.SDKException:
                # Just forget it if there's any trouble
                pass

        AttachmentsColumnWithCache = functools.partial(
            AttachmentsColumn, server_cache=server_cache
        )

        return (
            column_headers,
            (
                utils.get_item_properties(
                    s,
                    columns,
                    formatters={
                        'metadata': format_columns.DictColumn,
                        'attachments': AttachmentsColumnWithCache,
                    },
                )
                for s in data
            ),
        )


class MigrateVolume(command.Command):
    _description = _("Migrate volume to a new host")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'volume',
            metavar="<volume>",
            help=_("Volume to migrate (name or ID)"),
        )
        destination_group = parser.add_mutually_exclusive_group(required=True)
        destination_group.add_argument(
            '--host',
            metavar="<host>",
            help=_(
                "Destination host (takes the form: host@backend-name#pool)"
            ),
        )
        destination_group.add_argument(
            '--cluster',
            metavar="<cluster>",
            help=_(
                "Destination cluster to migrate the volume to "
                "(requires --os-volume-api-version 3.16 or higher)"
            ),
        )
        parser.add_argument(
            '--force-host-copy',
            action="store_true",
            help=_(
                "Enable generic host-based force-migration, "
                "which bypasses driver optimizations"
            ),
        )
        parser.add_argument(
            '--lock-volume',
            action="store_true",
            help=_(
                "If specified, the volume state will be locked "
                "and will not allow a migration to be aborted "
                "(possibly by another operation)"
            ),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )
        volume = volume_client.find_volume(
            parsed_args.volume, ignore_missing=False
        )

        if parsed_args.cluster and not sdk_utils.supports_microversion(
            volume_client, '3.16'
        ):
            msg = _(
                "--os-volume-api-version 3.16 or greater is required to "
                "support the volume migration with cluster"
            )
            raise exceptions.CommandError(msg)

        volume_client.migrate_volume(
            volume.id,
            host=parsed_args.host,
            force_host_copy=parsed_args.force_host_copy,
            lock_volume=parsed_args.lock_volume,
            cluster=parsed_args.cluster,
        )


class SetVolume(command.Command):
    _description = _("Set volume properties")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'volume',
            metavar='<volume>',
            help=_('Volume to modify (name or ID)'),
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            help=_('New volume name'),
        )
        parser.add_argument(
            '--size',
            metavar='<size>',
            type=int,
            help=_('Extend volume size in GB'),
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_('New volume description'),
        )
        parser.add_argument(
            "--no-property",
            dest="no_property",
            action="store_true",
            help=_(
                "Remove all properties from <volume> "
                "(specify both --no-property and --property to "
                "remove the current properties before setting "
                "new properties.)"
            ),
        )
        parser.add_argument(
            '--property',
            metavar='<key=value>',
            action=parseractions.KeyValueAction,
            dest='properties',
            help=_(
                'Set a property on this volume '
                '(repeat option to set multiple properties)'
            ),
        )
        parser.add_argument(
            '--image-property',
            metavar='<key=value>',
            action=parseractions.KeyValueAction,
            dest='image_properties',
            help=_(
                'Set an image property on this volume '
                '(repeat option to set multiple image properties)'
            ),
        )
        parser.add_argument(
            "--state",
            metavar="<state>",
            choices=[
                'available',
                'error',
                'creating',
                'deleting',
                'in-use',
                'attaching',
                'detaching',
                'error_deleting',
                'maintenance',
            ],
            help=_(
                'New volume state ("available", "error", "creating", '
                '"deleting", "in-use", "attaching", "detaching", '
                '"error_deleting" or "maintenance") (admin only) '
                '(This option simply changes the state of the volume '
                'in the database with no regard to actual status, '
                'exercise caution when using)'
            ),
        )
        attached_group = parser.add_mutually_exclusive_group()
        attached_group.add_argument(
            "--attached",
            action="store_true",
            help=_(
                'Set volume attachment status to "attached" '
                '(admin only) '
                '(This option simply changes the state of the volume '
                'in the database with no regard to actual status, '
                'exercise caution when using)'
            ),
        )
        attached_group.add_argument(
            "--detached",
            action="store_true",
            help=_(
                'Set volume attachment status to "detached" '
                '(admin only) '
                '(This option simply changes the state of the volume '
                'in the database with no regard to actual status, '
                'exercise caution when using)'
            ),
        )
        parser.add_argument(
            '--type',
            metavar='<volume-type>',
            help=_('New volume type (name or ID)'),
        )
        parser.add_argument(
            '--retype-policy',
            metavar='<retype-policy>',
            choices=['never', 'on-demand'],
            help=argparse.SUPPRESS,
        )
        parser.add_argument(
            '--migration-policy',
            metavar='<migration-policy>',
            choices=['never', 'on-demand'],
            help=_(
                'Migration policy while re-typing volume '
                '("never" or "on-demand", default is "never" ) '
                '(available only when --type option is specified)'
            ),
        )
        bootable_group = parser.add_mutually_exclusive_group()
        bootable_group.add_argument(
            "--bootable",
            action="store_true",
            dest="bootable",
            default=None,
            help=_("Mark volume as bootable"),
        )
        bootable_group.add_argument(
            "--non-bootable",
            action="store_false",
            dest="bootable",
            default=None,
            help=_("Mark volume as non-bootable"),
        )
        readonly_group = parser.add_mutually_exclusive_group()
        readonly_group.add_argument(
            "--read-only",
            action="store_true",
            dest="read_only",
            default=None,
            help=_("Set volume to read-only access mode"),
        )
        readonly_group.add_argument(
            "--read-write",
            action="store_false",
            dest="read_only",
            default=None,
            help=_("Set volume to read-write access mode"),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )
        volume = volume_client.find_volume(
            parsed_args.volume, ignore_missing=False
        )

        result = 0
        if parsed_args.retype_policy:
            msg = _(
                "The '--retype-policy' option has been deprecated in favor "
                "of '--migration-policy' option. The '--retype-policy' option "
                "will be removed in a future release. Please use "
                "'--migration-policy' instead."
            )
            self.log.warning(msg)

        if parsed_args.size:
            try:
                if parsed_args.size <= volume.size:
                    msg = (
                        _("New size must be greater than %s GB") % volume.size
                    )
                    raise exceptions.CommandError(msg)
                if volume.status not in ('available', 'in-use'):
                    msg = (
                        _(
                            "Volume is in %s state, it must be available "
                            "or in-use before size can be extended."
                        )
                        % volume.status
                    )
                    raise exceptions.CommandError(msg)
                if (
                    volume.status == 'in-use'
                    and not sdk_utils.supports_microversion(
                        volume_client, '3.42'
                    )
                ):
                    msg = _(
                        "--os-volume-api-version 3.42 or greater is "
                        "required to extend in-use volumes."
                    )
                    raise exceptions.CommandError(msg)
                volume_client.extend_volume(volume, parsed_args.size)
            except Exception as e:
                LOG.error(_("Failed to set volume size: %s"), e)
                result += 1

        if parsed_args.no_property:
            try:
                volume_client.delete_volume_metadata(volume)
            except Exception as e:
                LOG.error(_("Failed to clean volume properties: %s"), e)
                result += 1

        if parsed_args.properties:
            try:
                volume_client.set_volume_metadata(
                    volume, **parsed_args.properties
                )
            except Exception as e:
                LOG.error(_("Failed to set volume properties: %s"), e)
                result += 1

        if parsed_args.image_properties:
            try:
                volume_client.set_volume_image_metadata(
                    volume, **parsed_args.image_properties
                )
            except Exception as e:
                LOG.error(_("Failed to set image properties: %s"), e)
                result += 1

        if parsed_args.state:
            try:
                volume_client.reset_volume_status(
                    volume, status=parsed_args.state
                )
            except Exception as e:
                LOG.error(_("Failed to set volume state: %s"), e)
                result += 1

        if parsed_args.attached:
            try:
                volume_client.reset_volume_status(
                    volume, attach_status="attached"
                )
            except Exception as e:
                LOG.error(_("Failed to set volume attach-status: %s"), e)
                result += 1

        if parsed_args.detached:
            try:
                volume_client.reset_volume_status(
                    volume, attach_status="detached"
                )
            except Exception as e:
                LOG.error(_("Failed to set volume attach-status: %s"), e)
                result += 1

        if parsed_args.bootable is not None:
            try:
                volume_client.set_volume_bootable_status(
                    volume, parsed_args.bootable
                )
            except Exception as e:
                LOG.error(_("Failed to set volume bootable property: %s"), e)
                result += 1

        if parsed_args.read_only is not None:
            try:
                volume_client.set_volume_readonly(
                    volume, parsed_args.read_only
                )
            except Exception as e:
                LOG.error(
                    _("Failed to set volume read-only access mode flag: %s"),
                    e,
                )
                result += 1

        policy = parsed_args.migration_policy or parsed_args.retype_policy
        if parsed_args.type:
            # get the migration policy
            migration_policy = 'never'
            if policy:
                migration_policy = policy
            try:
                # find the volume type
                volume_type = volume_client.find_type(
                    parsed_args.type, ignore_missing=False
                )
                # reset to the new volume type
                volume_client.retype_volume(
                    volume, volume_type, migration_policy
                )
            except Exception as e:
                LOG.error(_("Failed to set volume type: %s"), e)
                result += 1
        elif policy:
            # If the "--migration-policy" is specified without "--type"
            LOG.warning(
                _("'%s' option will not work without '--type' option"),
                (
                    '--migration-policy'
                    if parsed_args.migration_policy
                    else '--retype-policy'
                ),
            )

        kwargs = {}
        if parsed_args.name:
            kwargs['name'] = parsed_args.name
        if parsed_args.description:
            kwargs['description'] = parsed_args.description
        if kwargs:
            try:
                volume_client.update_volume(volume, **kwargs)
            except Exception as e:
                LOG.error(
                    _(
                        "Failed to update volume display name "
                        "or display description: %s"
                    ),
                    e,
                )
                result += 1

        if result > 0:
            raise exceptions.CommandError(
                _("One or more of the set operations failed")
            )


class ShowVolume(command.ShowOne):
    _description = _("Display volume details")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'volume',
            metavar="<volume>",
            help=_("Volume to display (name or ID)"),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )
        volume = volume_client.find_volume(
            parsed_args.volume, ignore_missing=False
        )

        data = _format_volume(volume)
        col_headers, col_data = zip(*sorted(data.items()))
        return col_headers, col_data


class UnsetVolume(command.Command):
    _description = _("Unset volume properties")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'volume',
            metavar='<volume>',
            help=_('Volume to modify (name or ID)'),
        )
        parser.add_argument(
            '--property',
            metavar='<key>',
            action='append',
            dest='properties',
            help=_(
                'Remove a property from volume '
                '(repeat option to remove multiple properties)'
            ),
        )
        parser.add_argument(
            '--image-property',
            metavar='<key>',
            action='append',
            dest='image_properties',
            help=_(
                'Remove an image property from volume '
                '(repeat option to remove multiple image properties)'
            ),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )
        volume = volume_client.find_volume(
            parsed_args.volume, ignore_missing=False
        )

        result = 0
        if parsed_args.properties:
            try:
                volume_client.delete_volume_metadata(
                    volume.id, keys=parsed_args.properties
                )
            except Exception as e:
                LOG.error(_("Failed to unset volume properties: %s"), e)
                result += 1

        if parsed_args.image_properties:
            try:
                volume_client.delete_volume_image_metadata(
                    volume.id, keys=parsed_args.image_properties
                )
            except Exception as e:
                LOG.error(_("Failed to unset image properties: %s"), e)
                result += 1

        if result > 0:
            raise exceptions.CommandError(
                _("One or more of the unset operations failed")
            )


class VolumeSummary(command.ShowOne):
    _description = _("Show a summary of all volumes in this deployment.")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--all-projects',
            action='store_true',
            default=False,
            help=_('Include all projects (admin only)'),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )

        if not sdk_utils.supports_microversion(volume_client, '3.12'):
            msg = _(
                "--os-volume-api-version 3.12 or greater is required to "
                "support the 'volume summary' command"
            )
            raise exceptions.CommandError(msg)

        columns = [
            'total_count',
            'total_size',
        ]
        column_headers = [
            'Total Count',
            'Total Size',
        ]
        if sdk_utils.supports_microversion(volume_client, '3.36'):
            columns.append('metadata')
            column_headers.append('Metadata')

        # set value of 'all_tenants' when using project option
        all_projects = parsed_args.all_projects

        vol_summary = volume_client.summary(all_projects)

        return (
            column_headers,
            utils.get_item_properties(
                vol_summary,
                columns,
                formatters={'metadata': format_columns.DictColumn},
            ),
        )


class VolumeRevertToSnapshot(command.Command):
    _description = _("Revert a volume to a snapshot.")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'snapshot',
            metavar="<snapshot>",
            help=_(
                'Name or ID of the snapshot to restore. The snapshot must '
                'be the most recent one known to cinder.'
            ),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )

        if not sdk_utils.supports_microversion(volume_client, '3.40'):
            msg = _(
                "--os-volume-api-version 3.40 or greater is required to "
                "support the 'volume revert snapshot' command"
            )
            raise exceptions.CommandError(msg)

        snapshot = volume_client.find_snapshot(
            parsed_args.snapshot,
            ignore_missing=False,
        )
        volume = volume_client.find_volume(
            snapshot.volume_id,
            ignore_missing=False,
        )

        volume_client.revert_volume_to_snapshot(volume, snapshot)



#   Copyright 2013 Nebula Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from unittest import mock

from keystoneauth1 import exceptions as ks_exc
from osc_lib.cli import format_columns
from osc_lib import exceptions
from osc_lib import utils

from openstackclient.identity.v2_0 import project
from openstackclient.tests.unit.identity.v2_0 import fakes as identity_fakes


class TestProject(identity_fakes.TestIdentity):
    fake_project = identity_fakes.FakeProject.create_one_project()
    fake_projects = identity_fakes.FakeProject.create_projects()

    columns = (
        'description',
        'enabled',
        'id',
        'name',
    )
    datalist = (
        fake_project.description,
        True,
        fake_project.id,
        fake_project.name,
    )
    datalists = (
        (
            fake_projects[0].description,
            True,
            fake_projects[0].id,
            fake_projects[0].name,
        ),
        (
            fake_projects[1].description,
            True,
            fake_projects[1].id,
            fake_projects[1].name,
        ),
    )

    def setUp(self):
        super().setUp()

        # Get a shortcut to the TenantManager Mock
        self.projects_mock = self.identity_client.tenants
        self.projects_mock.reset_mock()


class TestProjectCreate(TestProject):
    def setUp(self):
        super().setUp()

        self.projects_mock.create.return_value = self.fake_project

        # Get the command object to test
        self.cmd = project.CreateProject(self.app, None)

    def test_project_create_no_options(self):
        arglist = [
            self.fake_project.name,
        ]
        verifylist = [
            ('enable', False),
            ('disable', False),
            ('name', self.fake_project.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'description': None,
            'enabled': True,
        }
        self.projects_mock.create.assert_called_with(
            self.fake_project.name, **kwargs
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, data)

    def test_project_create_description(self):
        arglist = [
            '--description',
            'new desc',
            self.fake_project.name,
        ]
        verifylist = [
            ('description', 'new desc'),
            ('name', self.fake_project.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'description': 'new desc',
            'enabled': True,
        }
        self.projects_mock.create.assert_called_with(
            self.fake_project.name, **kwargs
        )

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, data)

    def test_project_create_enable(self):
        arglist = [
            '--enable',
            self.fake_project.name,
        ]
        verifylist = [
            ('enable', True),
            ('disable', False),
            ('name', self.fake_project.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'description': None,
            'enabled': True,
        }
        self.projects_mock.create.assert_called_with(
            self.fake_project.name, **kwargs
        )

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, data)

    def test_project_create_disable(self):
        arglist = [
            '--disable',
            self.fake_project.name,
        ]
        verifylist = [
            ('enable', False),
            ('disable', True),
            ('name', self.fake_project.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'description': None,
            'enabled': False,
        }
        self.projects_mock.create.assert_called_with(
            self.fake_project.name, **kwargs
        )

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, data)

    def test_project_create_property(self):
        arglist = [
            '--property',
            'fee=fi',
            '--property',
            'fo=fum',
            self.fake_project.name,
        ]
        verifylist = [
            ('properties', {'fee': 'fi', 'fo': 'fum'}),
            ('name', self.fake_project.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'description': None,
            'enabled': True,
            'fee': 'fi',
            'fo': 'fum',
        }
        self.projects_mock.create.assert_called_with(
            self.fake_project.name, **kwargs
        )

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, data)

    def test_project_create_or_show_exists(self):
        def _raise_conflict(*args, **kwargs):
            raise ks_exc.Conflict(None)

        # need to make this throw an exception...
        self.projects_mock.create.side_effect = _raise_conflict

        self.projects_mock.get.return_value = self.fake_project

        arglist = [
            '--or-show',
            self.fake_project.name,
        ]
        verifylist = [
            ('or_show', True),
            ('name', self.fake_project.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # ProjectManager.create(name, description, enabled)
        self.projects_mock.get.assert_called_with(self.fake_project.name)

        # Set expected values
        kwargs = {
            'description': None,
            'enabled': True,
        }
        self.projects_mock.create.assert_called_with(
            self.fake_project.name, **kwargs
        )

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, data)

    def test_project_create_or_show_not_exists(self):
        arglist = [
            '--or-show',
            self.fake_project.name,
        ]
        verifylist = [
            ('or_show', True),
            ('name', self.fake_project.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'description': None,
            'enabled': True,
        }
        self.projects_mock.create.assert_called_with(
            self.fake_project.name, **kwargs
        )

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, data)


class TestProjectDelete(TestProject):
    def setUp(self):
        super().setUp()

        # This is the return value for utils.find_resource()
        self.projects_mock.get.return_value = self.fake_project
        self.projects_mock.delete.return_value = None

        # Get the command object to test
        self.cmd = project.DeleteProject(self.app, None)

    def test_project_delete_no_options(self):
        arglist = [
            self.fake_project.id,
        ]
        verifylist = [
            ('projects', [self.fake_project.id]),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.projects_mock.delete.assert_called_with(
            self.fake_project.id,
        )
        self.assertIsNone(result)

    @mock.patch.object(utils, 'find_resource')
    def test_delete_multi_projects_with_exception(self, find_mock):
        find_mock.side_effect = [self.fake_project, exceptions.CommandError]
        arglist = [
            self.fake_project.id,
            'unexist_project',
        ]
        verifylist = [
            ('projects', arglist),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        try:
            self.cmd.take_action(parsed_args)
            self.fail('CommandError should be raised.')
        except exceptions.CommandError as e:
            self.assertEqual('1 of 2 projects failed to delete.', str(e))

        find_mock.assert_any_call(self.projects_mock, self.fake_project.id)
        find_mock.assert_any_call(self.projects_mock, 'unexist_project')

        self.assertEqual(2, find_mock.call_count)
        self.projects_mock.delete.assert_called_once_with(self.fake_project.id)


class TestProjectList(TestProject):
    def setUp(self):
        super().setUp()

        self.projects_mock.list.return_value = [self.fake_project]

        # Get the command object to test
        self.cmd = project.ListProject(self.app, None)

    def test_project_list_no_options(self):
        arglist = []
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)
        self.projects_mock.list.assert_called_with()

        collist = ('ID', 'Name')
        self.assertEqual(collist, columns)
        datalist = (
            (
                self.fake_project.id,
                self.fake_project.name,
            ),
        )
        self.assertEqual(datalist, tuple(data))

    def test_project_list_long(self):
        arglist = [
            '--long',
        ]
        verifylist = [
            ('long', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)
        self.projects_mock.list.assert_called_with()

        collist = ('ID', 'Name', 'Description', 'Enabled')
        self.assertEqual(collist, columns)
        datalist = (
            (
                self.fake_project.id,
                self.fake_project.name,
                self.fake_project.description,
                True,
            ),
        )
        self.assertEqual(datalist, tuple(data))

    def test_project_list_sort(self):
        self.projects_mock.list.return_value = self.fake_projects

        arglist = [
            '--sort',
            'name:asc',
        ]
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        (columns, data) = self.cmd.take_action(parsed_args)
        self.projects_mock.list.assert_called_with()

        collist = ('ID', 'Name')
        self.assertEqual(collist, columns)

        if self.fake_projects[0].name > self.fake_projects[1].name:
            datalists = (
                (self.fake_projects[1].id, self.fake_projects[1].name),
                (self.fake_projects[0].id, self.fake_projects[0].name),
            )
        else:
            datalists = (
                (self.fake_projects[0].id, self.fake_projects[0].name),
                (self.fake_projects[1].id, self.fake_projects[1].name),
            )

        self.assertEqual(datalists, tuple(data))


class TestProjectSet(TestProject):
    def setUp(self):
        super().setUp()

        self.projects_mock.get.return_value = self.fake_project
        self.projects_mock.update.return_value = self.fake_project

        # Get the command object to test
        self.cmd = project.SetProject(self.app, None)

    def test_project_set_no_options(self):
        arglist = [
            self.fake_project.name,
        ]
        verifylist = [
            ('project', self.fake_project.name),
            ('enable', False),
            ('disable', False),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.assertIsNone(result)

    def test_project_set_unexist_project(self):
        arglist = [
            "unexist-project",
        ]
        verifylist = [
            ('project', "unexist-project"),
            ('name', None),
            ('description', None),
            ('enable', False),
            ('disable', False),
            ('properties', None),
        ]
        self.projects_mock.get.side_effect = exceptions.NotFound(None)
        self.projects_mock.find.side_effect = exceptions.NotFound(None)

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )

    def test_project_set_name(self):
        arglist = [
            '--name',
            self.fake_project.name,
            self.fake_project.name,
        ]
        verifylist = [
            ('name', self.fake_project.name),
            ('enable', False),
            ('disable', False),
            ('project', self.fake_project.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'description': self.fake_project.description,
            'enabled': True,
            'tenant_name': self.fake_project.name,
        }
        self.projects_mock.update.assert_called_with(
            self.fake_project.id, **kwargs
        )
        self.assertIsNone(result)

    def test_project_set_description(self):
        arglist = [
            '--description',
            self.fake_project.description,
            self.fake_project.name,
        ]
        verifylist = [
            ('description', self.fake_project.description),
            ('enable', False),
            ('disable', False),
            ('project', self.fake_project.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'description': self.fake_project.description,
            'enabled': True,
            'tenant_name': self.fake_project.name,
        }
        self.projects_mock.update.assert_called_with(
            self.fake_project.id, **kwargs
        )
        self.assertIsNone(result)

    def test_project_set_enable(self):
        arglist = [
            '--enable',
            self.fake_project.name,
        ]
        verifylist = [
            ('enable', True),
            ('disable', False),
            ('project', self.fake_project.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'description': self.fake_project.description,
            'enabled': True,
            'tenant_name': self.fake_project.name,
        }
        self.projects_mock.update.assert_called_with(
            self.fake_project.id, **kwargs
        )
        self.assertIsNone(result)

    def test_project_set_disable(self):
        arglist = [
            '--disable',
            self.fake_project.name,
        ]
        verifylist = [
            ('enable', False),
            ('disable', True),
            ('project', self.fake_project.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'description': self.fake_project.description,
            'enabled': False,
            'tenant_name': self.fake_project.name,
        }
        self.projects_mock.update.assert_called_with(
            self.fake_project.id, **kwargs
        )
        self.assertIsNone(result)

    def test_project_set_property(self):
        arglist = [
            '--property',
            'fee=fi',
            '--property',
            'fo=fum',
            self.fake_project.name,
        ]
        verifylist = [
            ('properties', {'fee': 'fi', 'fo': 'fum'}),
            ('project', self.fake_project.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'description': self.fake_project.description,
            'enabled': True,
            'tenant_name': self.fake_project.name,
            'fee': 'fi',
            'fo': 'fum',
        }
        self.projects_mock.update.assert_called_with(
            self.fake_project.id, **kwargs
        )
        self.assertIsNone(result)


class TestProjectShow(TestProject):
    fake_proj_show = identity_fakes.FakeProject.create_one_project()

    def setUp(self):
        super().setUp()

        self.projects_mock.get.return_value = self.fake_proj_show

        # Get the command object to test
        self.cmd = project.ShowProject(self.app, None)

    def test_project_show(self):
        arglist = [
            self.fake_proj_show.id,
        ]
        verifylist = [
            ('project', self.fake_proj_show.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)
        self.projects_mock.get.assert_called_with(
            self.fake_proj_show.id,
        )

        collist = ('description', 'enabled', 'id', 'name', 'properties')
        self.assertEqual(collist, columns)
        datalist = (
            self.fake_proj_show.description,
            True,
            self.fake_proj_show.id,
            self.fake_proj_show.name,
            format_columns.DictColumn({}),
        )
        self.assertCountEqual(datalist, data)


class TestProjectUnset(TestProject):
    attr = {'fee': 'fi', 'fo': 'fum'}
    fake_proj = identity_fakes.FakeProject.create_one_project(attr)

    def setUp(self):
        super().setUp()

        self.projects_mock.get.return_value = self.fake_proj

        # Get the command object to test
        self.cmd = project.UnsetProject(self.app, None)

    def test_project_unset_no_options(self):
        arglist = [
            self.fake_proj.name,
        ]
        verifylist = [
            ('project', self.fake_proj.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.assertIsNone(result)

    def test_project_unset_key(self):
        arglist = [
            '--property',
            'fee',
            '--property',
            'fo',
            self.fake_proj.name,
        ]
        verifylist = [
            ('properties', ['fee', 'fo']),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        # Set expected values
        kwargs = {
            'description': self.fake_proj.description,
            'enabled': True,
            'fee': None,
            'fo': None,
            'id': self.fake_proj.id,
            'name': self.fake_proj.name,
        }

        self.projects_mock.update.assert_called_with(
            self.fake_proj.id, **kwargs
        )
        self.assertIsNone(result)



#   Copyright 2013 Nebula Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from unittest import mock

from openstackclient.api import object_store_v1 as object_store
from openstackclient.tests.unit import utils


ACCOUNT_ID = 'tqbfjotld'
ENDPOINT = 'https://0.0.0.0:6482/v1/' + ACCOUNT_ID

container_name = 'bit-bucket'
container_bytes = 1024
container_count = 1

container_name_2 = 'archive'
container_name_3 = 'bit-blit'

CONTAINER = {
    'name': container_name,
    'bytes': container_bytes,
    'count': container_count,
}

CONTAINER_2 = {
    'name': container_name_2,
    'bytes': container_bytes * 2,
    'count': container_count * 2,
}

CONTAINER_3 = {
    'name': container_name_3,
    'bytes': container_bytes * 3,
    'count': container_count * 3,
}

object_name_1 = 'punch-card'
object_bytes_1 = 80
object_hash_1 = '1234567890'
object_content_type_1 = 'text'
object_modified_1 = 'today'

object_name_2 = 'floppy-disk'
object_bytes_2 = 1440000
object_hash_2 = '0987654321'
object_content_type_2 = 'text'
object_modified_2 = 'today'

OBJECT = {
    'name': object_name_1,
    'bytes': object_bytes_1,
    'hash': object_hash_1,
    'content_type': object_content_type_1,
    'last_modified': object_modified_1,
}

object_1_content = b'object 1 content'

OBJECT_2 = {
    'name': object_name_2,
    'bytes': object_bytes_2,
    'hash': object_hash_2,
    'content_type': object_content_type_2,
    'last_modified': object_modified_2,
}

object_upload_name = 'test-object-name'


class FakeClientMixin:
    def setUp(self):
        super().setUp()

        self.app.client_manager.object_store = mock.Mock(
            spec=object_store.APIv1
        )
        self.object_store_client = self.app.client_manager.object_store


class TestObjectV1(FakeClientMixin, utils.TestCommand): ...



#   Copyright 2012-2013 OpenStack Foundation
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Identity v3 Credential action implementations"""

import argparse
from collections.abc import Iterable, Sequence
import logging
from typing import Any

from openstack.identity.v3 import credential as _credential
from openstack import utils as sdk_utils
from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.i18n import _
from openstackclient.identity import common

LOG = logging.getLogger(__name__)


def _format_credential(
    credential: _credential.Credential,
) -> tuple[tuple[str, ...], Iterable[Any]]:
    columns = (
        'blob',
        'id',
        'project_id',
        'type',
        'user_id',
    )
    return (
        columns,
        utils.get_item_properties(
            credential,
            columns,
        ),
    )


class CreateCredential(command.ShowOne):
    _description = _("Create new credential")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'user',
            metavar='<user>',
            help=_('user that owns the credential (name or ID)'),
        )
        parser.add_argument(
            '--type',
            default="cert",
            metavar='<type>',
            help=_('New credential type: cert, ec2, totp and so on'),
        )
        parser.add_argument(
            'data',
            metavar='<data>',
            help=_('New credential data'),
        )
        parser.add_argument(
            '--project',
            metavar='<project>',
            help=_(
                'Project which limits the scope of the credential (name or ID)'
            ),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        identity_client = sdk_utils.ensure_service_version(
            self.app.client_manager.sdk_connection.identity, '3'
        )
        user_id = identity_client.find_user(
            parsed_args.user, ignore_missing=False
        ).id
        if parsed_args.project:
            project = identity_client.find_project(
                parsed_args.project, ignore_missing=False
            ).id
        else:
            project = None
        credential = identity_client.create_credential(
            user_id=user_id,
            type=parsed_args.type,
            blob=parsed_args.data,
            project_id=project,
        )

        return _format_credential(credential)


class DeleteCredential(command.Command):
    _description = _("Delete credential(s)")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'credential',
            metavar='<credential-id>',
            nargs='+',
            help=_('ID of credential(s) to delete'),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        identity_client = sdk_utils.ensure_service_version(
            self.app.client_manager.sdk_connection.identity, '3'
        )
        result = 0
        for i in parsed_args.credential:
            try:
                identity_client.delete_credential(i)
            except Exception as e:
                result += 1
                LOG.error(
                    _(
                        "Failed to delete credentials with "
                        "ID '%(credential)s': %(e)s"
                    ),
                    {'credential': i, 'e': e},
                )

        if result > 0:
            total = len(parsed_args.credential)
            msg = _("%(result)s of %(total)s credential failed to delete.") % {
                'result': result,
                'total': total,
            }
            raise exceptions.CommandError(msg)


class ListCredential(command.Lister):
    _description = _("List credentials")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--user',
            metavar='<user>',
            help=_('Filter credentials by <user> (name or ID)'),
        )
        common.add_user_domain_option_to_parser(parser)
        parser.add_argument(
            '--type',
            metavar='<type>',
            help=_('Filter credentials by type: cert, ec2, totp and so on'),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[tuple[str, ...], Iterable[tuple[Any, ...]]]:
        identity_client = sdk_utils.ensure_service_version(
            self.app.client_manager.sdk_connection.identity, '3'
        )

        kwargs = {}
        if parsed_args.user:
            domain_id = None
            if parsed_args.user_domain:
                domain_id = identity_client.find_domain(
                    parsed_args.user_domain, ignore_missing=False
                )
            user_id = identity_client.find_user(
                parsed_args.user, domain_id=domain_id, ignore_missing=False
            ).id
            kwargs["user_id"] = user_id

        if parsed_args.type:
            kwargs["type"] = parsed_args.type

        columns = ('ID', 'Type', 'User ID', 'Blob', 'Project ID')
        column_headers = ('ID', 'Type', 'User ID', 'Data', 'Project ID')
        data = identity_client.credentials(**kwargs)

        return (
            column_headers,
            (
                utils.get_item_properties(
                    s,
                    columns,
                    formatters={},
                )
                for s in data
            ),
        )


class SetCredential(command.Command):
    _description = _("Set credential properties")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'credential',
            metavar='<credential-id>',
            help=_('ID of credential to change'),
        )
        parser.add_argument(
            '--user',
            metavar='<user>',
            required=True,
            help=_('User that owns the credential (name or ID)'),
        )
        parser.add_argument(
            '--type',
            metavar='<type>',
            required=True,
            help=_('New credential type: cert, ec2, totp and so on'),
        )
        parser.add_argument(
            '--data',
            metavar='<data>',
            required=True,
            help=_('New credential data'),
        )
        parser.add_argument(
            '--project',
            metavar='<project>',
            help=_(
                'Project which limits the scope of the credential (name or ID)'
            ),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        identity_client = sdk_utils.ensure_service_version(
            self.app.client_manager.sdk_connection.identity, '3'
        )

        user_id = identity_client.find_user(
            parsed_args.user, ignore_missing=False
        ).id

        if parsed_args.project:
            project = identity_client.find_project(
                parsed_args.project, ignore_missing=False
            ).id
        else:
            project = None

        identity_client.update_credential(
            parsed_args.credential,
            user=user_id,
            type=parsed_args.type,
            blob=parsed_args.data,
            project=project,
        )


class ShowCredential(command.ShowOne):
    _description = _("Display credential details")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'credential',
            metavar='<credential-id>',
            help=_('ID of credential to display'),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        identity_client = sdk_utils.ensure_service_version(
            self.app.client_manager.sdk_connection.identity, '3'
        )
        credential = identity_client.get_credential(parsed_args.credential)

        return _format_credential(credential)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import argparse
from collections.abc import Iterable, Sequence
import logging
from typing import Any

from manilaclient import api_versions
from manilaclient.common.apiclient import utils as apiutils
from osc_lib import exceptions
from osc_lib import utils as oscutils

from openstackclient import command
from openstackclient.i18n import _

LOG = logging.getLogger(__name__)

MESSAGE_ATTRIBUTES = [
    'id',
    'resource_type',
    'resource_id',
    'action_id',
    'user_message',
    'message_level',
    'detail_id',
    'created_at',
    'expires_at',
    'request_id',
]


class DeleteMessage(command.Command):
    """Remove one or more messages."""

    _description = _("Remove one or more messages")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'message',
            metavar='<message>',
            nargs='+',
            help=_('ID of the message(s).'),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        share_client = self.app.client_manager.share
        failure_count = 0

        for message in parsed_args.message:
            try:
                message_ref = apiutils.find_resource(
                    share_client.messages, message
                )
                share_client.messages.delete(message_ref)
            except Exception as e:
                failure_count += 1
                LOG.error(
                    _("Delete for message %(message)s failed: %(e)s"),
                    {'message': message, 'e': e},
                )

        if failure_count > 0:
            raise exceptions.CommandError(
                _("Unable to delete some or all of the specified messages.")
            )


class ListMessage(command.Lister):
    """Lists all messages."""

    _description = _("Lists all messages")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--resource-id',
            metavar='<resource-id>',
            default=None,
            help=_('Filters results by a resource uuid. Default=None.'),
        )
        parser.add_argument(
            '--resource-type',
            metavar='<resource-type>',
            default=None,
            help=_(
                'Filters results by a resource type. Default=None. '
                'Example: "openstack message list --resource-type share"'
            ),
        )
        parser.add_argument(
            '--action-id',
            metavar='<action-id>',
            default=None,
            help=_('Filters results by action id. Default=None.'),
        )
        parser.add_argument(
            '--detail-id',
            metavar='<detail-id>',
            default=None,
            help=_('Filters results by detail id. Default=None.'),
        )
        parser.add_argument(
            '--request-id',
            metavar='<request-id>',
            default=None,
            help=_('Filters results by request id. Default=None.'),
        )
        parser.add_argument(
            '--message-level',
            metavar='<message-level>',
            default=None,
            help=_(
                'Filters results by the message level. Default=None. '
                'Example: "openstack message list --message-level ERROR".'
            ),
        )
        parser.add_argument(
            '--limit',
            metavar='<limit>',
            type=int,
            default=None,
            help=_('Maximum number of messages to return. (Default=None)'),
        )
        parser.add_argument(
            '--since',
            metavar='<since>',
            default=None,
            help=_(
                'Return only user messages created since given date. '
                'The date format must be conforming to ISO8601. '
                'Available only for microversion >= 2.52.'
            ),
        )
        parser.add_argument(
            '--before',
            metavar='<before>',
            default=None,
            help=_(
                'Return only user messages created before given date. '
                'The date format must be conforming to ISO8601. '
                'Available only for microversion >= 2.52.'
            ),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[tuple[Any, ...]]]:
        share_client = self.app.client_manager.share

        search_opts = {
            'limit': parsed_args.limit,
            'request_id': parsed_args.request_id,
            'resource_type': parsed_args.resource_type,
            'resource_id': parsed_args.resource_id,
            'action_id': parsed_args.action_id,
            'detail_id': parsed_args.detail_id,
            'message_level': parsed_args.message_level,
        }

        if share_client.api_version < api_versions.APIVersion("2.52"):
            if getattr(parsed_args, 'since') or getattr(parsed_args, 'before'):
                raise exceptions.CommandError(
                    _(
                        "Filtering messages by 'since' and 'before'"
                        " is possible only with Manila API version >=2.52"
                    )
                )
        else:
            search_opts['created_since'] = parsed_args.since
            search_opts['created_before'] = parsed_args.before

        messages = share_client.messages.list(search_opts=search_opts)
        columns = [
            'ID',
            'Resource Type',
            'Resource ID',
            'Action ID',
            'User Message',
            'Detail ID',
            'Created At',
        ]

        return (
            columns,
            (oscutils.get_item_properties(m, columns) for m in messages),
        )


class ShowMessage(command.ShowOne):
    """Show details about a message."""

    _description = _("Show details about a message")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'message', metavar='<message>', help=_('ID of the message.')
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], tuple[Any, ...]]:
        share_client = self.app.client_manager.share

        message = apiutils.find_resource(
            share_client.messages, parsed_args.message
        )

        return (
            MESSAGE_ATTRIBUTES,
            oscutils.get_dict_properties(message._info, MESSAGE_ATTRIBUTES),
        )



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import uuid

from openstackclient.tests.functional.network.v2 import common


class NetworkTests(common.NetworkTagTests):
    """Functional tests for network"""

    base_command = 'network'

    def test_network_create_compute(self):
        """Test Nova-net create options, delete"""
        if self.haz_network:
            self.skipTest("Skip Nova-net test")

        # Network create with minimum options
        name1 = uuid.uuid4().hex
        cmd_output = self.openstack(
            'network create ' + '--subnet 1.2.3.4/28 ' + name1,
            parse_output=True,
        )
        self.addCleanup(self.openstack, 'network delete ' + name1)
        self.assertIsNotNone(cmd_output["id"])

        self.assertEqual(
            name1,
            cmd_output["label"],
        )
        self.assertEqual(
            '1.2.3.0/28',
            cmd_output["cidr"],
        )

        # Network create with more options
        name2 = uuid.uuid4().hex
        cmd_output = self.openstack(
            'network create ' + '--subnet 1.2.4.4/28 ' + '--share ' + name2,
            parse_output=True,
        )
        self.addCleanup(self.openstack, 'network delete ' + name2)
        self.assertIsNotNone(cmd_output["id"])

        self.assertEqual(
            name2,
            cmd_output["label"],
        )
        self.assertEqual(
            '1.2.4.0/28',
            cmd_output["cidr"],
        )
        self.assertTrue(
            cmd_output["share_address"],
        )

    def test_network_create_network(self):
        """Test Neutron create options, delete"""
        if not self.haz_network:
            self.skipTest("No Network service present")

        # Get project IDs
        cmd_output = self.openstack('token issue ', parse_output=True)
        auth_project_id = cmd_output['project_id']

        cmd_output = self.openstack('project list ', parse_output=True)
        admin_project_id = None
        demo_project_id = None
        for p in cmd_output:
            if p['Name'] == 'admin':
                admin_project_id = p['ID']
            if p['Name'] == 'demo':
                demo_project_id = p['ID']

        # Verify assumptions:
        # * admin and demo projects are present
        # * demo and admin are distinct projects
        # * tests run as admin
        self.assertIsNotNone(admin_project_id)
        self.assertIsNotNone(demo_project_id)
        self.assertNotEqual(admin_project_id, demo_project_id)
        self.assertEqual(admin_project_id, auth_project_id)

        # Network create with no options
        name1 = uuid.uuid4().hex
        cmd_output = self.openstack(
            'network create ' + name1,
            parse_output=True,
        )
        self.addCleanup(self.openstack, 'network delete ' + name1)
        self.assertIsNotNone(cmd_output["id"])

        # Check the default values
        self.assertEqual(
            admin_project_id,
            cmd_output["project_id"],
        )
        self.assertEqual(
            '',
            cmd_output["description"],
        )
        self.assertEqual(
            True,
            cmd_output["admin_state_up"],
        )
        self.assertFalse(
            cmd_output["shared"],
        )
        self.assertEqual(
            False,
            cmd_output["router:external"],
        )

        # Network create with options
        name2 = uuid.uuid4().hex
        cmd_output = self.openstack(
            'network create ' + '--project demo ' + name2,
            parse_output=True,
        )
        self.addCleanup(self.openstack, 'network delete ' + name2)
        self.assertIsNotNone(cmd_output["id"])
        self.assertEqual(
            demo_project_id,
            cmd_output["project_id"],
        )
        self.assertEqual(
            '',
            cmd_output["description"],
        )

    def test_network_delete_compute(self):
        """Test create, delete multiple"""
        if self.haz_network:
            self.skipTest("Skip Nova-net test")

        name1 = uuid.uuid4().hex
        cmd_output = self.openstack(
            'network create ' + '--subnet 9.8.7.6/28 ' + name1,
            parse_output=True,
        )
        self.assertIsNotNone(cmd_output["id"])
        self.assertEqual(
            name1,
            cmd_output["label"],
        )

        name2 = uuid.uuid4().hex
        cmd_output = self.openstack(
            'network create ' + '--subnet 8.7.6.5/28 ' + name2,
            parse_output=True,
        )
        self.assertIsNotNone(cmd_output["id"])
        self.assertEqual(
            name2,
            cmd_output["label"],
        )

    def test_network_delete_network(self):
        """Test create, delete multiple"""
        if not self.haz_network:
            self.skipTest("No Network service present")

        name1 = uuid.uuid4().hex
        cmd_output = self.openstack(
            'network create ' + '--description aaaa ' + name1,
            parse_output=True,
        )
        self.assertIsNotNone(cmd_output["id"])
        self.assertEqual(
            'aaaa',
            cmd_output["description"],
        )

        name2 = uuid.uuid4().hex
        cmd_output = self.openstack(
            'network create ' + '--description bbbb ' + name2,
            parse_output=True,
        )
        self.assertIsNotNone(cmd_output["id"])
        self.assertEqual(
            'bbbb',
            cmd_output["description"],
        )

        del_output = self.openstack(f'network delete {name1} {name2}')
        self.assertOutput('', del_output)

    def test_network_list(self):
        """Test create defaults, list filters, delete"""
        name1 = uuid.uuid4().hex
        if self.haz_network:
            network_options = '--description aaaa --no-default '
        else:
            network_options = '--subnet 3.4.5.6/28 '
        cmd_output = self.openstack(
            'network create ' + network_options + name1,
            parse_output=True,
        )
        self.addCleanup(self.openstack, f'network delete {name1}')
        self.assertIsNotNone(cmd_output["id"])
        if self.haz_network:
            self.assertEqual(
                'aaaa',
                cmd_output["description"],
            )
            # Check the default values
            self.assertEqual(
                True,
                cmd_output["admin_state_up"],
            )
            self.assertFalse(cmd_output["shared"])
            self.assertEqual(
                False,
                cmd_output["router:external"],
            )
            self.assertFalse(cmd_output["is_default"])
            self.assertTrue(cmd_output["port_security_enabled"])
        else:
            self.assertEqual(
                '3.4.5.0/28',
                cmd_output["cidr"],
            )

        name2 = uuid.uuid4().hex
        if self.haz_network:
            network_options = '--description bbbb --disable '
        else:
            network_options = '--subnet 4.5.6.7/28 '
        cmd_output = self.openstack(
            f'network create --share {network_options}{name2}',
            parse_output=True,
        )
        self.addCleanup(self.openstack, 'network delete ' + name2)
        self.assertIsNotNone(cmd_output["id"])
        if self.haz_network:
            self.assertEqual(
                'bbbb',
                cmd_output["description"],
            )
            self.assertEqual(
                False,
                cmd_output["admin_state_up"],
            )
            self.assertTrue(cmd_output["shared"])
            self.assertFalse(cmd_output["is_default"])
            self.assertTrue(cmd_output["port_security_enabled"])
        else:
            self.assertEqual(
                '4.5.6.0/28',
                cmd_output["cidr"],
            )
            self.assertTrue(cmd_output["share_address"])

        # Test list
        cmd_output = self.openstack(
            "network list ",
            parse_output=True,
        )
        col_name = [x["Name"] for x in cmd_output]
        self.assertIn(name1, col_name)
        self.assertIn(name2, col_name)

        # Test list --long
        if self.haz_network:
            cmd_output = self.openstack(
                "network list --long",
                parse_output=True,
            )
            col_name = [x["Name"] for x in cmd_output]
            self.assertIn(name1, col_name)
            self.assertIn(name2, col_name)

        # Test list --long --enable
        if self.haz_network:
            cmd_output = self.openstack(
                "network list --enable --long",
                parse_output=True,
            )
            col_name = [x["Name"] for x in cmd_output]
            self.assertIn(name1, col_name)
            self.assertNotIn(name2, col_name)

        # Test list --long --disable
        if self.haz_network:
            cmd_output = self.openstack(
                "network list --disable --long",
                parse_output=True,
            )
            col_name = [x["Name"] for x in cmd_output]
            self.assertNotIn(name1, col_name)
            self.assertIn(name2, col_name)

        # Test list --share
        if self.haz_network:
            cmd_output = self.openstack(
                "network list --share ",
                parse_output=True,
            )
            col_name = [x["Name"] for x in cmd_output]
            self.assertNotIn(name1, col_name)
            self.assertIn(name2, col_name)

        # Test list --no-share
        if self.haz_network:
            cmd_output = self.openstack(
                "network list --no-share ",
                parse_output=True,
            )
            col_name = [x["Name"] for x in cmd_output]
            self.assertIn(name1, col_name)
            self.assertNotIn(name2, col_name)

    def test_network_dhcp_agent(self):
        if not self.haz_network:
            self.skipTest("No Network service present")

        if not self.is_extension_enabled("agent"):
            self.skipTest("No dhcp_agent_scheduler extension present")

        if not self.is_extension_enabled("dhcp_agent_scheduler"):
            self.skipTest("No dhcp_agent_scheduler extension present")

        # Get DHCP Agent ID
        cmd_output = self.openstack(
            'network agent list --agent-type dhcp',
            parse_output=True,
        )
        if not cmd_output:
            self.skipTest("No agents with type=dhcp available")

        agent_id = cmd_output[0]['ID']

        name1 = uuid.uuid4().hex
        cmd_output = self.openstack(
            f'network create --description aaaa {name1}',
            parse_output=True,
        )

        self.addCleanup(self.openstack, f'network delete {name1}')

        # Get network ID
        network_id = cmd_output['id']

        # Add Agent to Network
        self.openstack(
            f'network agent add network --dhcp {agent_id} {network_id}'
        )

        # Test network list --agent
        cmd_output = self.openstack(
            f'network list --agent {agent_id}',
            parse_output=True,
        )

        # Cleanup
        # Remove Agent from Network
        self.openstack(
            f'network agent remove network --dhcp {agent_id} {network_id}'
        )

        # Assert
        col_name = [x["ID"] for x in cmd_output]
        self.assertIn(network_id, col_name)

    def test_network_set(self):
        """Tests create options, set, show, delete"""
        if not self.haz_network:
            self.skipTest("No Network service present")

        name = uuid.uuid4().hex
        cmd_output = self.openstack(
            'network create '
            '--description aaaa '
            '--enable '
            '--no-share '
            '--internal '
            '--no-default '
            f'--enable-port-security {name}',
            parse_output=True,
        )
        self.addCleanup(self.openstack, f'network delete {name}')
        self.assertIsNotNone(cmd_output["id"])
        self.assertEqual(
            'aaaa',
            cmd_output["description"],
        )
        self.assertEqual(
            True,
            cmd_output["admin_state_up"],
        )
        self.assertFalse(cmd_output["shared"])
        self.assertEqual(
            False,
            cmd_output["router:external"],
        )

        self.assertFalse(cmd_output["is_default"])
        self.assertTrue(cmd_output["port_security_enabled"])

        raw_output = self.openstack(
            'network set '
            '--description cccc '
            '--disable '
            '--share '
            '--external '
            f'--disable-port-security {name}'
        )
        self.assertOutput('', raw_output)

        cmd_output = self.openstack(
            'network show ' + name,
            parse_output=True,
        )

        self.assertEqual(
            'cccc',
            cmd_output["description"],
        )
        self.assertEqual(
            False,
            cmd_output["admin_state_up"],
        )
        self.assertTrue(cmd_output["shared"])
        self.assertEqual(
            True,
            cmd_output["router:external"],
        )
        self.assertFalse(cmd_output["is_default"])
        self.assertFalse(cmd_output["port_security_enabled"])



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import logging
from unittest import mock
import uuid

from manilaclient import api_versions
from osc_lib import exceptions
from osc_lib import utils as oscutils

from openstackclient.share.v2 import share_group_snapshots
from openstackclient.tests.unit.share.v2 import fakes as share_fakes
from openstackclient.tests.unit import utils as test_utils

LOG = logging.getLogger(__name__)


class TestShareGroupSnapshot(share_fakes.TestShare):
    def setUp(self):
        super().setUp()

        self.groups_mock = self.share_client.share_groups
        self.groups_mock.reset_mock()

        self.group_snapshot_mocks = self.share_client.share_group_snapshots
        self.group_snapshot_mocks.reset_mock()

        self.set_share_api_version(api_versions.MAX_VERSION)


class TestCreateShareGroupSnapshot(TestShareGroupSnapshot):
    def setUp(self):
        super().setUp()

        self.share_group = share_fakes.FakeShareGroup.create_one_share_group()
        self.groups_mock.get.return_value = self.share_group

        self.share_group_snapshot = share_fakes.FakeShareGroupSnapshot.create_one_share_group_snapshot()
        self.group_snapshot_mocks.create.return_value = (
            self.share_group_snapshot
        )
        self.group_snapshot_mocks.get.return_value = self.share_group_snapshot

        self.cmd = share_group_snapshots.CreateShareGroupSnapshot(
            self.app, None
        )

        self.data = tuple(self.share_group_snapshot._info.values())
        self.columns = tuple(self.share_group_snapshot._info.keys())

    def test_share_group_snapshot_create_missing_args(self):
        arglist = []
        verifylist = []

        self.assertRaises(
            test_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_share_group_snapshot_create(self):
        arglist = [self.share_group.id]
        verifylist = [('share_group', self.share_group.id)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.group_snapshot_mocks.create.assert_called_with(
            self.share_group, name=None, description=None
        )

        self.assertCountEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_share_group_snapshot_create_options(self):
        arglist = [
            self.share_group.id,
            '--name',
            self.share_group_snapshot.name,
            '--description',
            self.share_group_snapshot.description,
        ]
        verifylist = [
            ('share_group', self.share_group.id),
            ('name', self.share_group_snapshot.name),
            ('description', self.share_group_snapshot.description),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.group_snapshot_mocks.create.assert_called_with(
            self.share_group,
            name=self.share_group_snapshot.name,
            description=self.share_group_snapshot.description,
        )

        self.assertCountEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_share_group_snapshot_create_wait(self):
        arglist = [self.share_group.id, '--wait']
        verifylist = [('share_group', self.share_group.id), ('wait', True)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        with mock.patch('osc_lib.utils.wait_for_status', return_value=True):
            self.group_snapshot_mocks.create.assert_called_with(
                self.share_group,
                name=None,
                description=None,
            )
            self.group_snapshot_mocks.get.assert_called_with(
                self.share_group_snapshot.id
            )
            self.assertCountEqual(self.columns, columns)
            self.assertCountEqual(self.data, data)

    @mock.patch('openstackclient.share.v2.share_group_snapshots.LOG')
    def test_share_group_snapshot_create_wait_exception(self, mock_logger):
        arglist = [self.share_group.id, '--wait']
        verifylist = [('share_group', self.share_group.id), ('wait', True)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        with mock.patch('osc_lib.utils.wait_for_status', return_value=False):
            columns, data = self.cmd.take_action(parsed_args)

            self.group_snapshot_mocks.create.assert_called_with(
                self.share_group,
                name=None,
                description=None,
            )

            mock_logger.error.assert_called_with(
                "ERROR: Share group snapshot is in error state."
            )

            self.group_snapshot_mocks.get.assert_called_with(
                self.share_group_snapshot.id
            )
            self.assertCountEqual(self.columns, columns)
            self.assertCountEqual(self.data, data)


class TestDeleteShareGroupSnapshot(TestShareGroupSnapshot):
    def setUp(self):
        super().setUp()

        self.share_group_snapshot = share_fakes.FakeShareGroupSnapshot.create_one_share_group_snapshot()
        self.group_snapshot_mocks.get.return_value = self.share_group_snapshot

        self.cmd = share_group_snapshots.DeleteShareGroupSnapshot(
            self.app, None
        )

    def test_share_group_snapshot_delete_missing_args(self):
        arglist = []
        verifylist = []

        self.assertRaises(
            test_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_share_group_snapshot_delete(self):
        arglist = [self.share_group_snapshot.id]
        verifylist = [('share_group_snapshot', [self.share_group_snapshot.id])]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.group_snapshot_mocks.delete.assert_called_with(
            self.share_group_snapshot, force=False
        )
        self.assertIsNone(result)

    def test_share_group_snapshot_delete_force(self):
        arglist = [self.share_group_snapshot.id, '--force']
        verifylist = [
            ('share_group_snapshot', [self.share_group_snapshot.id]),
            ('force', True),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.group_snapshot_mocks.delete.assert_called_with(
            self.share_group_snapshot, force=True
        )
        self.assertIsNone(result)

    def test_share_group_snapshot_delete_multiple(self):
        share_group_snapshots = (
            share_fakes.FakeShareGroupSnapshot.create_share_group_snapshots(
                count=2
            )
        )
        arglist = [share_group_snapshots[0].id, share_group_snapshots[1].id]
        verifylist = [
            (
                'share_group_snapshot',
                [share_group_snapshots[0].id, (share_group_snapshots[1].id)],
            )
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.assertEqual(
            self.group_snapshot_mocks.delete.call_count,
            len(share_group_snapshots),
        )
        self.assertIsNone(result)

    def test_share_group_snapshot_delete_exception(self):
        arglist = [self.share_group_snapshot.id]
        verifylist = [('share_group_snapshot', [self.share_group_snapshot.id])]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.group_snapshot_mocks.delete.side_effect = (
            exceptions.CommandError()
        )
        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )

    def test_share_group_snapshot_delete_wait(self):
        arglist = [self.share_group_snapshot.id, '--wait']
        verifylist = [
            ('share_group_snapshot', [self.share_group_snapshot.id]),
            ('wait', True),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        with mock.patch('osc_lib.utils.wait_for_delete', return_value=True):
            result = self.cmd.take_action(parsed_args)

            self.group_snapshot_mocks.delete.assert_called_with(
                self.share_group_snapshot, force=False
            )
            self.group_snapshot_mocks.get.assert_called_with(
                self.share_group_snapshot.id
            )
            self.assertIsNone(result)

    def test_share_group_snapshot_delete_wait_exception(self):
        arglist = [self.share_group_snapshot.id, '--wait']
        verifylist = [
            ('share_group_snapshot', [self.share_group_snapshot.id]),
            ('wait', True),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        with mock.patch('osc_lib.utils.wait_for_delete', return_value=False):
            self.assertRaises(
                exceptions.CommandError, self.cmd.take_action, parsed_args
            )


class TestShowShareGroupSnapshot(TestShareGroupSnapshot):
    def setUp(self):
        super().setUp()

        self.share_group_snapshot = share_fakes.FakeShareGroupSnapshot.create_one_share_group_snapshot()
        self.group_snapshot_mocks.get.return_value = self.share_group_snapshot

        self.cmd = share_group_snapshots.ShowShareGroupSnapshot(self.app, None)

        self.data = tuple(self.share_group_snapshot._info.values())
        self.columns = tuple(self.share_group_snapshot._info.keys())

    def test_share_group_snapshot_show_missing_args(self):
        arglist = []
        verifylist = []

        self.assertRaises(
            test_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_share_group_show(self):
        arglist = [self.share_group_snapshot.id]
        verifylist = [('share_group_snapshot', self.share_group_snapshot.id)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.group_snapshot_mocks.get.assert_called_with(
            self.share_group_snapshot.id
        )

        self.assertCountEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)


class TestSetShareGroupSnapshot(TestShareGroupSnapshot):
    def setUp(self):
        super().setUp()

        self.share_group_snapshot = share_fakes.FakeShareGroupSnapshot.create_one_share_group_snapshot()
        self.group_snapshot_mocks.get.return_value = self.share_group_snapshot

        self.cmd = share_group_snapshots.SetShareGroupSnapshot(self.app, None)

        self.data = tuple(self.share_group_snapshot._info.values())
        self.columns = tuple(self.share_group_snapshot._info.keys())

    def test_set_share_group_snapshot_name_description(self):
        group_snapshot_name = 'group-snapshot-name-' + uuid.uuid4().hex
        group_snapshot_description = (
            'group-snapshot-description-' + uuid.uuid4().hex
        )
        arglist = [
            self.share_group_snapshot.id,
            '--name',
            group_snapshot_name,
            '--description',
            group_snapshot_description,
        ]
        verifylist = [
            ('share_group_snapshot', self.share_group_snapshot.id),
            ('name', group_snapshot_name),
            ('description', group_snapshot_description),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.group_snapshot_mocks.update.assert_called_with(
            self.share_group_snapshot,
            name=parsed_args.name,
            description=parsed_args.description,
        )
        self.assertIsNone(result)

    def test_set_share_group_snapshot_status(self):
        arglist = [self.share_group_snapshot.id, '--status', 'available']
        verifylist = [
            ('share_group_snapshot', self.share_group_snapshot.id),
            ('status', 'available'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.group_snapshot_mocks.reset_state.assert_called_with(
            self.share_group_snapshot, 'available'
        )
        self.assertIsNone(result)

    def test_set_share_group_snapshot_exception(self):
        arglist = [self.share_group_snapshot.id, '--status', 'available']
        verifylist = [
            ('share_group_snapshot', self.share_group_snapshot.id),
            ('status', 'available'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.group_snapshot_mocks.reset_state.side_effect = Exception()

        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )


class TestUnsetShareGroupSnapshot(TestShareGroupSnapshot):
    def setUp(self):
        super().setUp()

        self.share_group_snapshot = share_fakes.FakeShareGroupSnapshot.create_one_share_group_snapshot()
        self.group_snapshot_mocks.get.return_value = self.share_group_snapshot

        self.cmd = share_group_snapshots.UnsetShareGroupSnapshot(
            self.app, None
        )

    def test_unset_share_group_snapshot_name_description(self):
        arglist = [self.share_group_snapshot.id, '--name', '--description']
        verifylist = [
            ('share_group_snapshot', self.share_group_snapshot.id),
            ('name', True),
            ('description', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.group_snapshot_mocks.update.assert_called_with(
            self.share_group_snapshot, name='', description=''
        )
        self.assertIsNone(result)

    def test_unset_share_group_snapshot_name_exception(self):
        arglist = [
            self.share_group_snapshot.id,
            '--name',
        ]
        verifylist = [
            ('share_group_snapshot', self.share_group_snapshot.id),
            ('name', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.group_snapshot_mocks.update.side_effect = Exception()

        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )


class TestListShareGroupSnapshot(TestShareGroupSnapshot):
    columns = [
        'ID',
        'Name',
        'Status',
        'Description',
    ]

    def setUp(self):
        super().setUp()

        self.share_group = share_fakes.FakeShareGroup.create_one_share_group()
        self.groups_mock.get.return_value = self.share_group

        self.share_group_snapshot = (
            share_fakes.FakeShareGroupSnapshot.create_one_share_group_snapshot(
                {'share_group_id': self.share_group.id}
            )
        )

        self.share_group_snapshots_list = [self.share_group_snapshot]
        self.group_snapshot_mocks.list.return_value = (
            self.share_group_snapshots_list
        )

        self.values = (
            oscutils.get_dict_properties(s._info, self.columns)
            for s in self.share_group_snapshots_list
        )

        self.cmd = share_group_snapshots.ListShareGroupSnapshot(self.app, None)

    def test_share_group_snapshot_list_no_options(self):
        arglist = []
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.group_snapshot_mocks.list.assert_called_once_with(
            search_opts={
                'all_tenants': False,
                'name': None,
                'status': None,
                'share_group_id': None,
                'limit': None,
                'offset': None,
            }
        )

        self.assertEqual(self.columns, columns)
        self.assertEqual(list(self.values), list(data))

    def test_share_group_snapshot_list_detail_all_projects(self):
        columns_detail = [
            'ID',
            'Name',
            'Status',
            'Description',
            'Created At',
            'Share Group ID',
            'Project ID',
        ]

        values = (
            oscutils.get_dict_properties(s._info, columns_detail)
            for s in self.share_group_snapshots_list
        )

        arglist = [
            '--detailed',
            '--all-projects',
        ]

        verifylist = [
            ('detailed', True),
            ('all_projects', True),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.group_snapshot_mocks.list.assert_called_once_with(
            search_opts={
                'all_tenants': True,
                'name': None,
                'status': None,
                'share_group_id': None,
                'limit': None,
                'offset': None,
            }
        )

        self.assertEqual(columns_detail, columns)
        self.assertEqual(list(values), list(data))

    def test_share_group_snapshot_list_search_options(self):
        arglist = [
            '--name',
            self.share_group_snapshot.name,
            '--status',
            self.share_group_snapshot.status,
            '--share-group',
            self.share_group.id,
        ]
        verifylist = [
            ('name', self.share_group_snapshot.name),
            ('status', self.share_group_snapshot.status),
            ('share_group', self.share_group.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.groups_mock.get.assert_called_with(self.share_group.id)
        self.group_snapshot_mocks.list.assert_called_once_with(
            search_opts={
                'all_tenants': False,
                'name': self.share_group_snapshot.name,
                'status': self.share_group_snapshot.status,
                'share_group_id': self.share_group.id,
                'limit': None,
                'offset': None,
            }
        )

        self.assertEqual(self.columns, columns)
        self.assertEqual(list(self.values), list(data))


class TestListShareGroupSnapshotMembers(TestShareGroupSnapshot):
    columns = [
        'Share ID',
        'Size',
    ]

    def setUp(self):
        super().setUp()

        self.share = share_fakes.FakeShare.create_one_share()

        self.share_group_snapshot = (
            share_fakes.FakeShareGroupSnapshot.create_one_share_group_snapshot(
                {
                    'members': [
                        {'share_id': self.share.id, 'size': self.share.size}
                    ]
                }
            )
        )

        self.group_snapshot_mocks.get.return_value = self.share_group_snapshot

        self.values = (
            oscutils.get_dict_properties(s, self.columns)
            for s in self.share_group_snapshot.members
        )

        self.cmd = share_group_snapshots.ListShareGroupSnapshotMembers(
            self.app, None
        )

    def test_share_group_snapshot_list_members(self):
        arglist = [self.share_group_snapshot.id]
        verifylist = [('share_group_snapshot', self.share_group_snapshot.id)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.group_snapshot_mocks.get.assert_called_with(
            self.share_group_snapshot.id
        )

        self.assertEqual(self.columns, columns)
        self.assertEqual(list(self.values), list(data))



# Copyright (c) 2016, Intel Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from openstackclient.network.v2 import network_qos_rule_type as _qos_rule_type
from openstackclient.tests.unit.network.v2 import fakes as network_fakes
from openstackclient.tests.unit import utils as tests_utils


class TestShowNetworkQosRuleType(network_fakes.TestNetworkV2):
    attrs = {'drivers': [{'name': 'driver 1', 'supported_parameters': []}]}
    # The QoS policies to show.
    qos_rule_type = network_fakes.create_one_qos_rule_type(attrs)
    columns = ('drivers', 'rule_type_name')
    columns = ('drivers', 'rule_type_name')
    data = [qos_rule_type.drivers, qos_rule_type.type]

    def setUp(self):
        super().setUp()
        self.network_client.get_qos_rule_type.return_value = self.qos_rule_type

        # Get the command object to test
        self.cmd = _qos_rule_type.ShowNetworkQosRuleType(self.app, None)

    def test_show_no_options(self):
        arglist = []
        verifylist = []

        # Missing required args should bail here
        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_show_all_options(self):
        arglist = [
            self.qos_rule_type.type,
        ]
        verifylist = [
            ('rule_type', self.qos_rule_type.type),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.get_qos_rule_type.assert_called_once_with(
            self.qos_rule_type.type
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(list(self.data), list(data))


class TestListNetworkQosRuleType(network_fakes.TestNetworkV2):
    # The QoS policies to list up.
    qos_rule_types = network_fakes.create_qos_rule_types(count=3)

    columns = ('Type',)
    data = []
    for qos_rule_type in qos_rule_types:
        data.append((qos_rule_type.type,))

    def setUp(self):
        super().setUp()
        self.network_client.qos_rule_types.return_value = self.qos_rule_types

        # Get the command object to test
        self.cmd = _qos_rule_type.ListNetworkQosRuleType(self.app, None)

    def test_qos_rule_type_list(self):
        arglist = []
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.qos_rule_types.assert_called_once_with(**{})
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, list(data))

    def test_qos_rule_type_list_all_supported(self):
        arglist = ['--all-supported']
        verifylist = [
            ('all_supported', True),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.qos_rule_types.assert_called_once_with(
            **{'all_supported': True}
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, list(data))

    def test_qos_rule_type_list_all_rules(self):
        arglist = ['--all-rules']
        verifylist = [
            ('all_rules', True),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.qos_rule_types.assert_called_once_with(
            **{'all_rules': True}
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, list(data))

    def test_qos_rule_type_list_pagination(self):
        arglist = [
            '--marker',
            self.qos_rule_types[0].type,
            '--limit',
            '1',
        ]
        verifylist = [
            ('marker', self.qos_rule_types[0].type),
            ('limit', 1),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.qos_rule_types.assert_called_once_with(
            **{
                'marker': self.qos_rule_types[0].type,
                'limit': 1,
            }
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, list(data))



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import argparse
from collections.abc import Iterable, Sequence
import logging
from typing import Any

from manilaclient import api_versions
from osc_lib.cli import parseractions
from osc_lib import exceptions
from osc_lib import utils as oscutils

from openstackclient import command
from openstackclient.i18n import _

LOG = logging.getLogger(__name__)


class CreateShareSecurityService(command.ShowOne):
    """Create security service used by project."""

    _description = _("Create security service used by project.")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'type',
            metavar='<type>',
            default=None,
            choices=['ldap', 'kerberos', 'active_directory'],
            help=_(
                "Security service type. Possible options are: "
                "'ldap', 'kerberos', 'active_directory'."
            ),
        )
        parser.add_argument(
            '--dns-ip',
            metavar='<dns-ip>',
            default=None,
            help=_(
                "DNS IP address of the security service used "
                "inside project's network."
            ),
        )
        parser.add_argument(
            '--ou',
            metavar='<ou>',
            default=None,
            help=_(
                "Security service OU (Organizational Unit). "
                "Available only for microversion >= 2.44."
            ),
        )
        parser.add_argument(
            '--server',
            metavar='<server>',
            default=None,
            help=_("Security service IP address or hostname."),
        )
        parser.add_argument(
            '--domain',
            metavar='<domain>',
            default=None,
            help=_("Security service domain."),
        )
        parser.add_argument(
            '--user',
            metavar='<user>',
            default=None,
            help=_("Security service user or group used by project."),
        )
        parser.add_argument(
            '--password',
            metavar='<password>',
            default=None,
            help=_("Password used by user."),
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            default=None,
            help=_("Security service name."),
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            default=None,
            help=_("Security service description."),
        )
        parser.add_argument(
            '--default-ad-site',
            metavar='<default_ad_site>',
            dest='default_ad_site',
            default=None,
            help=_(
                "Default AD site. Available only for "
                "microversion >= 2.76. Can be provided in the "
                "place of '--server' but not along with it."
            ),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[tuple[str, ...], tuple[Any, ...]]:
        share_client = self.app.client_manager.share

        kwargs = {
            'dns_ip': parsed_args.dns_ip,
            'server': parsed_args.server,
            'domain': parsed_args.domain,
            'user': parsed_args.user,
            'password': parsed_args.password,
            'name': parsed_args.name,
            'description': parsed_args.description,
        }

        if share_client.api_version >= api_versions.APIVersion("2.44"):
            kwargs['ou'] = parsed_args.ou
        elif parsed_args.ou:
            raise exceptions.CommandError(
                "Defining a security service Organizational Unit is "
                "available only for microversion >= 2.44"
            )

        if share_client.api_version >= api_versions.APIVersion("2.76"):
            kwargs['default_ad_site'] = parsed_args.default_ad_site
        elif parsed_args.default_ad_site:
            raise exceptions.CommandError(
                "Defining a security service Default AD site is "
                "available only for microversion >= 2.76"
            )

        if parsed_args.type == 'active_directory':
            server = parsed_args.server
            default_ad_site = parsed_args.default_ad_site
            if server and default_ad_site:
                raise exceptions.CommandError(
                    "Cannot create security service because both "
                    "server and 'default_ad_site' were provided. "
                    "Specify either server or 'default_ad_site'."
                )

        security_service = share_client.security_services.create(
            parsed_args.type, **kwargs
        )

        return self.dict2columns(security_service._info)


class DeleteShareSecurityService(command.Command):
    """Delete one or more security services."""

    _description = _("Delete one or more security services.")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'security_service',
            metavar='<security-service>',
            nargs="+",
            help=_("Name or ID of the security service(s) to delete."),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        share_client = self.app.client_manager.share
        result = 0

        for security_service in parsed_args.security_service:
            try:
                security_service_obj = oscutils.find_resource(
                    share_client.security_services, security_service
                )
                share_client.security_services.delete(security_service_obj)

            except Exception as e:
                result += 1
                LOG.error(
                    "Failed to delete security service with name or ID %s: %s",
                    security_service,
                    e,
                )

        if result > 0:
            total = len(parsed_args.security_service)
            msg = (
                f"{result} of {total} security services failed to be deleted."
            )
            raise exceptions.CommandError(msg)


class ShowShareSecurityService(command.ShowOne):
    """Show security service."""

    _description = _("Show security service.")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'security_service',
            metavar='<security-service>',
            help=_("Security service name or ID to show."),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[tuple[str, ...], tuple[Any, ...]]:
        share_client = self.app.client_manager.share

        security_service = oscutils.find_resource(
            share_client.security_services, parsed_args.security_service
        )

        data = security_service._info
        if parsed_args.formatter == 'table':
            if 'share_networks' in data.keys():
                data['share_networks'] = "\n".join(data['share_networks'])

        return self.dict2columns(data)


class SetShareSecurityService(command.Command):
    """Set security service."""

    _description = _("Set security service.")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'security_service',
            metavar='<security-service>',
            help=_("Security service name or ID."),
        )
        parser.add_argument(
            '--dns-ip',
            metavar='<dns-ip>',
            default=None,
            help=_("Set DNS IP address used inside project's network."),
        )
        parser.add_argument(
            '--ou',
            metavar='<ou>',
            default=None,
            help=_(
                "Set security service OU (Organizational Unit). "
                "Available only for microversion >= 2.44."
            ),
        )
        parser.add_argument(
            '--server',
            metavar='<server>',
            default=None,
            help=_("Set security service IP address or hostname."),
        )
        parser.add_argument(
            '--domain',
            metavar='<domain>',
            default=None,
            help=_("Set security service domain."),
        )
        parser.add_argument(
            '--user',
            metavar='<user>',
            default=None,
            help=_("Set security service user or group used by project."),
        )
        parser.add_argument(
            '--password',
            metavar='<password>',
            default=None,
            help=_("Set password used by user."),
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            default=None,
            help=_("Set security service name."),
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            default=None,
            help=_("Set security service description."),
        )
        parser.add_argument(
            '--default-ad-site',
            metavar='<default_ad_site>',
            dest='default_ad_site',
            default=None,
            help=_(
                "Default AD site. Available only for microversion >= 2.76."
            ),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        share_client = self.app.client_manager.share

        security_service = oscutils.find_resource(
            share_client.security_services, parsed_args.security_service
        )

        kwargs = {
            'dns_ip': parsed_args.dns_ip,
            'server': parsed_args.server,
            'domain': parsed_args.domain,
            'user': parsed_args.user,
            'password': parsed_args.password,
            'name': parsed_args.name,
            'description': parsed_args.description,
        }

        if share_client.api_version >= api_versions.APIVersion("2.44"):
            kwargs['ou'] = parsed_args.ou
        elif parsed_args.ou:
            raise exceptions.CommandError(
                _(
                    "Setting a security service Organizational Unit is "
                    "available only for microversion >= 2.44"
                )
            )

        if share_client.api_version >= api_versions.APIVersion("2.76"):
            kwargs['default_ad_site'] = parsed_args.default_ad_site
        elif parsed_args.default_ad_site:
            raise exceptions.CommandError(
                "Defining a security service Default AD site is "
                "available only for microversion >= 2.76"
            )

        if security_service.type == 'active_directory':
            server = parsed_args.server
            default_ad_site = parsed_args.default_ad_site
            if server and default_ad_site:
                raise exceptions.CommandError(
                    "Cannot set security service because both "
                    "server and 'default_ad_site' were provided. "
                    "Specify either server or 'default_ad_site'."
                )
        try:
            security_service.update(**kwargs)
        except Exception as e:
            raise exceptions.CommandError(
                f"One or more set operations failed: {e}"
            )


class UnsetShareSecurityService(command.Command):
    """Unset security service."""

    _description = _("Unset security service.")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'security_service',
            metavar='<security-service>',
            help=_("Security service name or ID."),
        )
        parser.add_argument(
            '--dns-ip',
            action='store_true',
            help=_("Unset DNS IP address used inside project's network."),
        )
        parser.add_argument(
            '--ou',
            action='store_true',
            help=_(
                "Unset security service OU (Organizational Unit). "
                "Available only for microversion >= 2.44."
            ),
        )
        parser.add_argument(
            '--server',
            action='store_true',
            help=_("Unset security service IP address or hostname."),
        )
        parser.add_argument(
            '--domain',
            action='store_true',
            help=_("Unset security service domain."),
        )
        parser.add_argument(
            '--user',
            action='store_true',
            help=_("Unset security service user or group used by project."),
        )
        parser.add_argument(
            '--password',
            action='store_true',
            help=_("Unset password used by user."),
        )
        parser.add_argument(
            '--name',
            action='store_true',
            help=_("Unset security service name."),
        )
        parser.add_argument(
            '--description',
            action='store_true',
            help=_("Unset security service description."),
        )
        parser.add_argument(
            '--default-ad-site',
            dest='default_ad_site',
            action='store_true',
            help=_(
                "Default AD site. Available only for microversion >= 2.76."
            ),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        share_client = self.app.client_manager.share

        security_service = oscutils.find_resource(
            share_client.security_services, parsed_args.security_service
        )

        kwargs: dict[str, str] = {}
        args = [
            'dns_ip',
            'server',
            'domain',
            'user',
            'password',
            'name',
            'description',
        ]
        for arg in args:
            if getattr(parsed_args, arg):
                # the SDK unsets a value if it is an empty string
                kwargs[arg] = ''

        if (
            parsed_args.ou
            and share_client.api_version >= api_versions.APIVersion("2.44")
        ):
            # the SDK unsets a value if it is an empty string
            kwargs['ou'] = ''

        elif parsed_args.ou:
            raise exceptions.CommandError(
                _(
                    "Unsetting a security service Organizational Unit is "
                    "available only for microversion >= 2.44"
                )
            )

        if (
            parsed_args.default_ad_site
            and share_client.api_version >= api_versions.APIVersion("2.76")
        ):
            # the SDK unsets a value if it is an empty string
            kwargs['default_ad_site'] = ''
        elif parsed_args.default_ad_site:
            raise exceptions.CommandError(
                _(
                    "Unsetting a security service Default AD site is "
                    "available only for microversion >= 2.76"
                )
            )

        try:
            security_service.update(**kwargs)
        except Exception as e:
            raise exceptions.CommandError(
                f"One or more unset operations failed: {e}"
            )


class ListShareSecurityService(command.Lister):
    """List security services."""

    _description = _("List security services.")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--all-projects',
            action='store_true',
            help=_("Display information from all projects (Admin only)."),
        )
        parser.add_argument(
            '--share-network',
            metavar='<share-network>',
            default=None,
            help=_("Filter results by share network name or ID."),
        )
        parser.add_argument(
            '--status',
            metavar='<status>',
            default=None,
            help=_("Filter results by status."),
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            default=None,
            help=_("Filter results by security service name."),
        )
        parser.add_argument(
            '--type',
            metavar='<type>',
            default=None,
            help=_("Filter results by security service type."),
        )
        parser.add_argument(
            '--user',
            metavar='<user>',
            default=None,
            help=_(
                "Filter results by security service user or group "
                "used by project."
            ),
        )
        parser.add_argument(
            '--dns-ip',
            metavar='<dns-ip>',
            default=None,
            help=_(
                "Filter results by DNS IP address used inside "
                "project's network."
            ),
        )
        parser.add_argument(
            '--ou',
            metavar='<ou>',
            default=None,
            help=_(
                "Filter results by security service OU "
                "(Organizational Unit). "
                "Available only for microversion >= 2.44."
            ),
        )
        parser.add_argument(
            '--default-ad-site',
            metavar='<default_ad_site>',
            dest='default_ad_site',
            default=None,
            help=_(
                "Filter results by security service default_ad_site. "
                "Available only for microversion >= 2.76."
            ),
        )
        parser.add_argument(
            '--server',
            metavar='<server>',
            default=None,
            help=_(
                "Filter results by security service IP address or hostname."
            ),
        )
        parser.add_argument(
            '--domain',
            metavar='<domain>',
            default=None,
            help=_("Filter results by security service domain."),
        )
        parser.add_argument(
            '--detail',
            action='store_true',
            help=_(
                "Show detailed information about filtered security services."
            ),
        )
        parser.add_argument(
            "--limit",
            metavar="<num-security-services>",
            type=int,
            default=None,
            action=parseractions.NonNegativeAction,
            help=_("Limit the number of security services returned"),
        )
        parser.add_argument(
            "--marker",
            metavar="<security-service>",
            help=_("The last security service ID of the previous page"),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[tuple[Any, ...]]]:
        share_client = self.app.client_manager.share

        columns = ['ID', 'Name', 'Status', 'Type']

        if parsed_args.all_projects:
            columns.append('Project ID')

        if parsed_args.detail:
            columns.append('Share Networks')

        search_opts = {
            'all_tenants': parsed_args.all_projects,
            'status': parsed_args.status,
            'name': parsed_args.name,
            'type': parsed_args.type,
            'user': parsed_args.user,
            'dns_ip': parsed_args.dns_ip,
            'server': parsed_args.server,
            'domain': parsed_args.domain,
            'offset': parsed_args.marker,
            'limit': parsed_args.limit,
        }

        if (
            parsed_args.ou
            and share_client.api_version >= api_versions.APIVersion("2.44")
        ):
            search_opts['ou'] = parsed_args.ou

        elif parsed_args.ou:
            raise exceptions.CommandError(
                _(
                    "Filtering results by security service Organizational "
                    "Unit is available only for microversion >= 2.44"
                )
            )

        if (
            parsed_args.default_ad_site
            and share_client.api_version >= api_versions.APIVersion("2.76")
        ):
            search_opts['default_ad_site'] = parsed_args.default_ad_site
        elif parsed_args.default_ad_site:
            raise exceptions.CommandError(
                _(
                    "Filtering results by security service Default AD site is "
                    "available only for microversion >= 2.76"
                )
            )

        if parsed_args.share_network:
            search_opts['share_network_id'] = oscutils.find_resource(
                share_client.share_networks, parsed_args.share_network
            ).id

        data = share_client.security_services.list(
            search_opts=search_opts, detailed=parsed_args.detail
        )

        return (
            columns,
            (oscutils.get_item_properties(s, columns) for s in data),
        )



#   Copyright 2017 Huawei, Inc. All rights reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Compute v2 Server operation event implementations"""

import argparse
from collections.abc import Iterable, Sequence
import logging
from typing import Any
import uuid

from cliff import columns
import iso8601
from openstack.compute.v2 import server_action as _server_action
from openstack import exceptions as sdk_exceptions
from openstack import utils as sdk_utils
from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.common import pagination
from openstackclient.i18n import _

LOG = logging.getLogger(__name__)


# TODO(stephenfin): Move this to osc_lib since it's useful elsewhere (e.g.
# glance)
def is_uuid_like(value: str) -> bool:
    """Returns validation of a value as a UUID.

    :param val: Value to verify
    :type val: string
    :returns: bool

    .. versionchanged:: 1.1.1
       Support non-lowercase UUIDs.
    """
    try:
        formatted_value = (
            value.replace('urn:', '')
            .replace('uuid:', '')
            .strip('{}')
            .replace('-', '')
            .lower()
        )
        return str(uuid.UUID(value)).replace('-', '') == formatted_value
    except (TypeError, ValueError, AttributeError):
        return False


class ServerActionEventColumn(columns.FormattableColumn[Any]):
    """Custom formatter for server action events.

    Format the :class:`~openstack.compute.v2.server_action.ServerActionEvent`
    objects as we'd like.
    """

    @staticmethod
    def _format_event(event: _server_action.ServerAction) -> dict[str, Any]:
        hidden_columns = ['id', 'name', 'location']
        _, columns = utils.get_osc_show_columns_for_sdk_resource(
            event, {}, hidden_columns
        )
        data = utils.get_item_properties(event, columns)
        return dict(zip(columns, data))

    def human_readable(self) -> str:
        events = [self._format_event(event) for event in self._value]
        return utils.format_list_of_dicts(events) or ""

    def machine_readable(self) -> Any:
        events = [self._format_event(event) for event in self._value]
        return events


def _get_server_event_columns(
    item: Any, client: Any
) -> tuple[tuple[str, ...], tuple[str, ...]]:
    hidden_columns = ['name', 'server_id', 'links', 'location']

    if not sdk_utils.supports_microversion(client, '2.58'):
        # updated_at was introduced in 2.58
        hidden_columns.append('updated_at')

    return utils.get_osc_show_columns_for_sdk_resource(
        item, {}, hidden_columns
    )


class ListServerEvent(command.Lister):
    """List recent events of a server.

    Specify ``--os-compute-api-version 2.21`` or higher to show events for a
    deleted server, specified by ID only.
    """

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'server',
            metavar='<server>',
            help=_('Server to list events (name or ID)'),
        )
        parser.add_argument(
            '--long',
            action='store_true',
            default=False,
            help=_("List additional fields in output"),
        )
        parser.add_argument(
            '--changes-since',
            dest='changes_since',
            metavar='<changes-since>',
            help=_(
                "List only server events changed later or equal to a certain "
                "point of time. The provided time should be an ISO 8061 "
                "formatted time, e.g. ``2016-03-04T06:27:59Z``. "
                "(supported with --os-compute-api-version 2.58 or above)"
            ),
        )
        parser.add_argument(
            '--changes-before',
            dest='changes_before',
            metavar='<changes-before>',
            help=_(
                "List only server events changed earlier or equal to a "
                "certain point of time. The provided time should be an ISO "
                "8061 formatted time, e.g. ``2016-03-04T06:27:59Z``. "
                "(supported with --os-compute-api-version 2.66 or above)"
            ),
        )
        pagination.add_marker_pagination_option_to_parser(parser)
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[tuple[str, ...], Iterable[tuple[Any, ...]]]:
        compute_client = self.app.client_manager.compute

        kwargs = {}

        if parsed_args.marker:
            if not sdk_utils.supports_microversion(compute_client, '2.58'):
                msg = _(
                    '--os-compute-api-version 2.58 or greater is required to '
                    'support the --marker option'
                )
                raise exceptions.CommandError(msg)

            kwargs['marker'] = parsed_args.marker

        if parsed_args.limit:
            if not sdk_utils.supports_microversion(compute_client, '2.58'):
                msg = _(
                    '--os-compute-api-version 2.58 or greater is required to '
                    'support the --limit option'
                )
                raise exceptions.CommandError(msg)

            kwargs['limit'] = parsed_args.limit
            kwargs['paginated'] = False

        if parsed_args.max_items is not None:
            kwargs['max_items'] = parsed_args.max_items

        if parsed_args.changes_since:
            if not sdk_utils.supports_microversion(compute_client, '2.58'):
                msg = _(
                    '--os-compute-api-version 2.58 or greater is required to '
                    'support the --changes-since option'
                )
                raise exceptions.CommandError(msg)

            try:
                iso8601.parse_date(parsed_args.changes_since)
            except (TypeError, iso8601.ParseError):
                msg = _('Invalid changes-since value: %s')
                raise exceptions.CommandError(msg % parsed_args.changes_since)

            kwargs['changes_since'] = parsed_args.changes_since

        if parsed_args.changes_before:
            if not sdk_utils.supports_microversion(compute_client, '2.66'):
                msg = _(
                    '--os-compute-api-version 2.66 or greater is required to '
                    'support the --changes-before option'
                )
                raise exceptions.CommandError(msg)

            try:
                iso8601.parse_date(parsed_args.changes_before)
            except (TypeError, iso8601.ParseError):
                msg = _('Invalid changes-before value: %s')
                raise exceptions.CommandError(msg % parsed_args.changes_before)

            kwargs['changes_before'] = parsed_args.changes_before

        try:
            server_id = compute_client.find_server(
                parsed_args.server, ignore_missing=False
            ).id
        except sdk_exceptions.ResourceNotFound:
            # If we fail to find the resource, it is possible the server is
            # deleted. Try once more using the <server> arg directly if it is a
            # UUID.
            if is_uuid_like(parsed_args.server):
                server_id = parsed_args.server
            else:
                raise

        data = compute_client.server_actions(server_id, **kwargs)

        columns: tuple[str, ...] = (
            'request_id',
            'server_id',
            'action',
            'start_time',
        )
        column_headers: tuple[str, ...] = (
            'Request ID',
            'Server ID',
            'Action',
            'Start Time',
        )

        if parsed_args.long:
            columns += (
                'message',
                'project_id',
                'user_id',
            )
            column_headers += (
                'Message',
                'Project ID',
                'User ID',
            )

        return (
            column_headers,
            (utils.get_item_properties(s, columns) for s in data),
        )


class ShowServerEvent(command.ShowOne):
    """Show server event details.

    Specify ``--os-compute-api-version 2.21`` or higher to show event details
    for a deleted server, specified by ID only. Specify
    ``--os-compute-api-version 2.51`` or higher to show event details for
    non-admin users.
    """

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'server',
            metavar='<server>',
            help=_('Server to show event details (name or ID)'),
        )
        parser.add_argument(
            'request_id',
            metavar='<request-id>',
            help=_('Request ID of the event to show (ID only)'),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        compute_client = self.app.client_manager.compute

        try:
            server_id = compute_client.find_server(
                parsed_args.server,
                ignore_missing=False,
            ).id
        except sdk_exceptions.ResourceNotFound:
            # If we fail to find the resource, it is possible the server is
            # deleted. Try once more using the <server> arg directly if it is a
            # UUID.
            if is_uuid_like(parsed_args.server):
                server_id = parsed_args.server
            else:
                raise

        server_action = compute_client.get_server_action(
            parsed_args.request_id,
            server_id,
        )
        # this is a bug is SDK
        assert server_action is not None

        column_headers, columns = _get_server_event_columns(
            server_action,
            compute_client,
        )

        return (
            column_headers,
            utils.get_item_properties(
                server_action,
                columns,
                formatters={'events': ServerActionEventColumn},
            ),
        )



#   Copyright 2012 OpenStack Foundation
#   Copyright 2013 Nebula Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Identity v2 EC2 Credentials action implementations"""

import argparse
from collections.abc import Iterable, Sequence
import logging
from typing import Any

from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.i18n import _


LOG = logging.getLogger(__name__)


class CreateEC2Creds(command.ShowOne):
    _description = _("Create EC2 credentials")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--project',
            metavar='<project>',
            help=_(
                'Create credentials in project '
                '(name or ID; default: current authenticated project)'
            ),
        )
        parser.add_argument(
            '--user',
            metavar='<user>',
            help=_(
                'Create credentials for user '
                '(name or ID; default: current authenticated user)'
            ),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        identity_client = self.app.client_manager.identity

        if parsed_args.project:
            project = utils.find_resource(
                identity_client.tenants,
                parsed_args.project,
            ).id
        else:
            # Get the project from the current auth
            project = self.app.client_manager.auth_ref.project_id
        if parsed_args.user:
            user = utils.find_resource(
                identity_client.users,
                parsed_args.user,
            ).id
        else:
            # Get the user from the current auth
            user = self.app.client_manager.auth_ref.user_id

        creds = identity_client.ec2.create(user, project)

        info = {}
        info.update(creds._info)

        if 'tenant_id' in info:
            info.update({'project_id': info.pop('tenant_id')})

        col_headers, col_data = zip(*sorted(info.items()))
        return col_headers, col_data


class DeleteEC2Creds(command.Command):
    _description = _("Delete EC2 credentials")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'access_keys',
            metavar='<access-key>',
            nargs='+',
            help=_('Credentials access key(s)'),
        )
        parser.add_argument(
            '--user',
            metavar='<user>',
            help=_('Delete credentials for user (name or ID)'),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        identity_client = self.app.client_manager.identity

        if parsed_args.user:
            user = utils.find_resource(
                identity_client.users,
                parsed_args.user,
            ).id
        else:
            # Get the user from the current auth
            user = self.app.client_manager.auth_ref.user_id

        result = 0
        for access_key in parsed_args.access_keys:
            try:
                identity_client.ec2.delete(user, access_key)
            except Exception as e:
                result += 1
                LOG.error(
                    _(
                        "Failed to delete EC2 credentials with "
                        "access key '%(access_key)s': %(e)s"
                    ),
                    {'access_key': access_key, 'e': e},
                )

        if result > 0:
            total = len(parsed_args.access_keys)
            msg = _("%(result)s of %(total)s EC2 keys failed to delete.") % {
                'result': result,
                'total': total,
            }
            raise exceptions.CommandError(msg)


class ListEC2Creds(command.Lister):
    _description = _("List EC2 credentials")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--user',
            metavar='<user>',
            help=_('Filter list by user (name or ID)'),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[tuple[str, ...], Iterable[tuple[Any, ...]]]:
        identity_client = self.app.client_manager.identity

        if parsed_args.user:
            user = utils.find_resource(
                identity_client.users,
                parsed_args.user,
            ).id
        else:
            # Get the user from the current auth
            user = self.app.client_manager.auth_ref.user_id

        columns = ('access', 'secret', 'tenant_id', 'user_id')
        column_headers = ('Access', 'Secret', 'Project ID', 'User ID')
        data = identity_client.ec2.list(user)

        return (
            column_headers,
            (
                utils.get_item_properties(
                    s,
                    columns,
                    formatters={},
                )
                for s in data
            ),
        )


class ShowEC2Creds(command.ShowOne):
    _description = _("Display EC2 credentials details")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'access_key',
            metavar='<access-key>',
            help=_('Credentials access key'),
        )
        parser.add_argument(
            '--user',
            metavar='<user>',
            help=_('Show credentials for user (name or ID)'),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        identity_client = self.app.client_manager.identity

        if parsed_args.user:
            user = utils.find_resource(
                identity_client.users,
                parsed_args.user,
            ).id
        else:
            # Get the user from the current auth
            user = self.app.client_manager.auth_ref.user_id

        creds = identity_client.ec2.get(user, parsed_args.access_key)

        info = {}
        info.update(creds._info)

        if 'tenant_id' in info:
            info.update({'project_id': info.pop('tenant_id')})

        col_headers, col_data = zip(*sorted(info.items()))
        return col_headers, col_data
